import collections
import heapq
import itertools
import json
import logging
import operator
import os
import random
import shutil
import subprocess
from ctypes import util as ctypes_util
from typing import (Any, Callable, Dict, FrozenSet, Iterator, List,
                    MutableMapping, Optional, Sequence, Set, Tuple, Union,
                    overload)

import cached_property
import pulp
from haoda import ir, util
from haoda.ir import arithmetic
from haoda.ir.arithmetic import base

import soda.visitor
from soda import grammar, mutator

RelativeAttr = int
AbsoluteAttr = int
Attr = Union[RelativeAttr, Tuple[RelativeAttr, Optional[AbsoluteAttr]]]
Index = Tuple[int, ...]

OrderedDict = collections.OrderedDict


class OrderedCounter(  # type: ignore
    collections.Counter, collections.OrderedDict):
  pass


_logger = logging.getLogger().getChild(__name__)


def extract_attr(node: ir.Node) -> Tuple[Tuple[int, ...], ir.Node]:
  """Extract attributes from a node.

  Extract relative and absolute attributes from a node. The relative attribute
  would be the load index and the absolute attribute is the normalized node.

  Args:
    node: The ir.node to be extracted.

  Returns:
    Tuple of rattr and aattr.
  """
  load = soda.visitor.get_load_set(node)[0]
  return load.idx, mutator.shift(node, load.idx)


def assemble_attr(rattr: Tuple[int, ...], aattr: ir.Node) -> ir.Node:
  """Assemble a node from attributes.

  The absolute attribute must be a normalized ir.Node. The relative attribute
  will be used as the shifting offset to obtain the original ir.Node.

  Args:
    rattr: The relative attribute.
    aattr: The absolute attribute.

  Returns:
    ir.Node assembled from the attributes.
  """
  return mutator.shift(aattr, rattr, op=operator.add)


class Linearizer:
  """Apply and restore linearization.

  This class stores the necessory information needed to apply and restore
  linearization. Instances of this class is callable.

  Attributes:
    maxs: List of integers, maximum index in each dimension.
    mins: List of integers, minimum index in each dimension.
    sizes: Tuple of integers, size of each linearized dimension.

  Properties:
    num_dim: Integer, number of dimensions.
    weights: List of integers, weight of each dimension.
    dims: Tuple of integers, all dimension indices.
  """

  def __init__(self,
               rattrs: Sequence[Sequence[int]],
               tile_size: Sequence[int] = ()):
    """Initialize the Linearizer with the given relative attribute tuples.

    Args:
      rattrs: Sequence of relative attributes. Each attribute is a sequence of
        integers.
    """
    num_dim = len(rattrs[0])
    self.maxs = [0] * num_dim
    self.mins = [0] * num_dim
    for d in self.dims:
      self.maxs[d] = max(rattr[d] for rattr in rattrs)
      self.mins[d] = min(rattr[d] for rattr in rattrs)
    if tile_size:
      self.sizes = tuple(tile_size)[:-1] + (
          (self.maxs[-1] - self.mins[-1] + 1) * 2 - 1,)
    else:
      self.sizes = tuple(
          (self.maxs[d] - self.mins[d] + 1) * 2 - 1 for d in self.dims)

  @property
  def num_dim(self) -> int:
    return len(self.maxs)

  @property
  def weights(self) -> List[int]:
    weights = [1] * self.num_dim
    for d in self.dims[1:]:
      weights[d] = weights[d - 1] * self.sizes[d - 1]
    return weights

  @property
  def dims(self) -> Tuple[int, ...]:
    return tuple(range(self.num_dim))

  def apply(self, rattr: Sequence[int]) -> int:
    return sum((rval - min_val) * weight
               for rval, weight, min_val in zip(rattr, self.weights, self.mins))

  def restore(self, rattr: int) -> Tuple[int, ...]:
    restored_attr = []  # type: List[int]
    for d in reversed(self.dims):
      rval = rattr // self.weights[d]
      rattr -= rval * self.weights[d]
      restored_attr.append(self.mins[d] + rval)
    return tuple(reversed(restored_attr))

  # pylint: disable=function-redefined
  @overload
  def __call__(self, rattr: Sequence[int]) -> int:
    ...

  # pylint: disable=function-redefined
  @overload
  def __call__(self, rattr: int) -> Tuple[int, ...]:
    ...

  def __call__(self, rattr):
    if isinstance(rattr, int):
      return self.restore(rattr)
    if isinstance(rattr, Sequence) and isinstance(rattr[0], int):
      return self.apply(rattr)
    raise TypeError('rattr needs to be an int or a Sequence of int')


def range_from_middle(n: int) -> Iterator[int]:
  """A range function that yields number from the middle to the sides.

  Args:
    n: Integer, the upper bound of the range.
  Yields:
    Integers, starting from the n / 2 towards 0 and n - 1.
  """
  middle = n // 2
  if n % 2 == 0:
    for shift in range(0, middle):
      yield middle - shift - 1
      yield middle + shift
  else:
    yield middle
    for shift in range(1, middle + 1):
      yield middle - shift
      yield middle + shift


def shuffle_range(n: int) -> Iterator[int]:
  lst = list(range(n))
  random.shuffle(lst)
  return iter(lst)


def set_optimizations(optimizations: Sequence[str]) -> None:
  Expression.set_optimizations(optimizations)


def computation_reuse(
    stencil: 'soda.core.Stencil'  # type: ignore
) -> 'soda.core.Stencil':  # type: ignore
  """Reuse computation by eliminating common subexpressions.

  The stencil object will be modified.

  Args:
    stencil: soda.core.Stencil object to work on.

  Returns:
    Modified stencil object.
  """
  method = stencil.optimizations.get('computation-reuse')
  if method is None or method == 'no':
    return stencil
  _logger.debug('invoke stencil computation reuse')

  # pylint: disable=unsubscriptable-object
  def visitor(node: ir.Node, cses: MutableMapping[ir.BinaryOp,
                                                  ir.Ref]) -> ir.Node:
    """Visitor for computation reuse.

    Args:
      csel: A dict mapping expressions to the new ir.Refs.

    Returns:
      Optimized ir.Node with computation reused.
    """
    try:
      expression = Expression(node, stencil)
      if expression.best_schedule is not None:
        _logger.debug('best schedule: (cost: %s)',
                      expression.best_schedule.cost)
        return expression.best_schedule.get_ir_node_with_cr(stencil, cses)
    except Expression.CannotHandle:
      pass
    return node

  new_local_stmts = []
  cses = OrderedDict()  # type: Dict[ir.BinaryOp, ir.Ref]
  seen = set()  # type: Set[ir.Expr]
  for stmt in itertools.chain(stencil.local_stmts, stencil.output_stmts):
    stmt.propagate_type(stencil.symbol_table)
    stmt.expr = stmt.expr.visit(visitor, cses)
    stmt.let = tuple(let.visit(visitor, cses) for let in stmt.let)
    for expr, ref in cses.items():
      if expr in seen:
        continue
      seen.add(expr)
      expr = stencil.propagate_type(expr, stmt)
      new_local_stmts.append(
          grammar.LocalStmt(ref=ref,
                            haoda_type=expr.haoda_type,
                            expr=expr,
                            let=stmt.let,
                            stencil=stencil))
      _logger.debug('computation reuse stmt: %s', new_local_stmts[-1])
  stencil.local_stmts.extend(new_local_stmts)

  # invalidate cached_property
  stencil.__dict__.pop('symbol_table', None)
  stencil.__dict__.pop('local_names', None)
  stencil.__dict__.pop('local_types', None)

  for stmt in itertools.chain(stencil.local_stmts, stencil.output_stmts):
    _logger.debug('simplify:   %s', stmt)
    stmt.expr = arithmetic.simplify(stmt.expr)
    stmt.let = arithmetic.simplify(stmt.let)
    _logger.debug('simplified: %s', stmt)
  _logger.info('stencil after CR: \n  %s', str(stencil).replace('\n', '\n  '))
  return stencil


class ScheduleBase:
  """Base class of Schedule and Schedules.

  Attributes:
    rattrs: Tuple of relative attributes.
    aattrs: Tuple of absolute attributes or None.
  """

  def __init__(self, rattrs: Tuple[RelativeAttr, ...],
               aattrs: Optional[Tuple[AbsoluteAttr, ...]]) -> None:
    self.rattrs = rattrs
    self.aattrs = aattrs

  def __getitem__(self,
                  key: int) -> Tuple[RelativeAttr, Optional[AbsoluteAttr]]:
    return self.rattrs[key], None if self.aattrs is None else self.aattrs[key]

  def __len__(self) -> int:
    return len(self.rattrs)

  def __iter__(self) -> Iterator[Tuple[RelativeAttr, Optional[AbsoluteAttr]]]:
    yield from zip(self.rattrs, self.aattrs or itertools.repeat(None))


class CommSchedule(ScheduleBase):
  """A schedule of an expression.

  A schedule represents a general schedule of n operands, described as a binary
  tree.

  Attributes:
    left: Left child of the schedule. It can be an integer if it is a leaf node,
        representing the index to the attributes; otherwise it is a
        CommSchedule.
    right: Right child of the schedule.
    distance: Distance between the left child and the right child.
  Properties:
    norm_attrs: Generator of all normalized attributes as Iterator[Attr].
    uniq_expr_dict: Unique expressions of this schedule as
        Dict[Tuple[Attr, ...], CommSchedule].
    uniq_expr_set: Unique expressions of this schedule as
        Set[Tuple[Attr, ...]].
    cost: Tuple of number of operations and total reuse distance required for
        this schedule.
    dependers: A dict mapping variable ids to its dependers set, which is an
        ordered dict mapping dependers to None.
    dependees: A dict mapping variable ids to its dependees dict, which maps
        dependees to a tuple of the first and last accessed offset.
  """

  def __init__(self,
               left: Union['CommSchedule', int, None],
               right: Union['CommSchedule', int, None],
               distance: RelativeAttr,
               rattrs: Tuple[RelativeAttr, ...],
               aattrs: Optional[Tuple[AbsoluteAttr, ...]] = None) -> None:
    self.left, self.right, self.distance = left, right, distance
    super().__init__(rattrs, aattrs)
    self._len = 1  # number of operations
    for child in left, right:
      if isinstance(child, CommSchedule):
        self._len += len(child)

  def __len__(self) -> int:
    """Number of operations."""
    return self._len

  def __lt__(self, rhs: 'CommSchedule') -> bool:
    return self.cost < rhs.cost

  def __eq__(self, other: object) -> bool:
    if not isinstance(other, CommSchedule):
      return NotImplemented
    return self.norm_attr_set == other.norm_attr_set

  def __hash__(self) -> int:
    return hash(self.norm_attr_set)

  def __str__(self) -> str:
    return self.to_str_with_offset(0)

  def to_str_with_offset(self, offset: int = 0) -> str:
    """Return the string representation assuming an offset.
    """
    if isinstance(self.left, CommSchedule):
      left = self.left.to_str_with_offset(offset)
    else:
      left = str(self.left)
    offset += self.distance
    if isinstance(self.right, CommSchedule):
      right = self.right.to_str_with_offset(offset)
    else:
      right = str(self.right)
    return '(%s==%s=>%s)' % (left, self.distance, right)

  def print_tree(self, printer=_logger.debug) -> None:
    base.print_tree(self.ir_node, printer)

  def bind_expression(self, expression: Optional['Expression']) \
      -> 'CommSchedule':
    """Bind an Expression to the schedule.
    """
    if expression is None:
      del self.aattrs_as_ir_nodes  # type: ignore
      del self.linearizer  # type: ignore
      del self.aattr_table  # type: ignore
      del self.operator  # type: ignore
    else:
      self.aattrs_as_ir_nodes = expression.aattrs_as_ir_nodes
      self.linearizer = expression.linearizer
      self.aattr_table = expression.aattr_table
      self.operator = expression.operator
    for child in self.left, self.right:
      if isinstance(child, CommSchedule):
        child.bind_expression(expression)
    return self

  @property
  def children(self) -> Iterator['CommSchedule']:
    if not hasattr(self, 'yielded_children'):
      self.yielded_children = []  # type: List[CommSchedule]
    yield from self.yielded_children
    yield from self.children_gen

  @cached_property.cached_property
  def children_gen(self) -> Iterator['CommSchedule']:
    self.yielded_children.append(self)
    yield self
    for child in self.left, self.right:
      if isinstance(child, CommSchedule):
        for schedule in child.children:
          self.yielded_children.append(schedule)
          yield schedule

  @cached_property.cached_property
  def cost(self) -> Tuple[int, int]:
    return self.num_ops, self.total_distance

  @cached_property.cached_property
  def num_ops(self) -> int:
    return len(set(self.children))

  def to_json(self):
    j = {'distance': self.distance}
    for child in 'left', 'right':
      j[child] = getattr(self, child)
      if isinstance(j[child], CommSchedule):
        j[child] = j[child].to_json()
    return j

  def _calc_dependency(self) -> None:
    """Calculate the dependency between reused variables.

    This function creates two dependency tables, i.e. dependers and dependees.
    """

    def get_attrs(schedule: CommSchedule,
                  reuses: MutableMapping[Schedule, int],
                  offset: Optional[int] = None) -> Iterator[Tuple[int, int]]:
      """Get all accesses of variables in a schedule.

      Args:
        schedule: The CommSchedule to work on.
        reuses: Dict mapping a reused CommSchedule to the variable id.
        offset: The global offset of the input schedule.

      Yields:
        Tuple of the offset and the variable id.
      """
      reused_vid = reuses.get(schedule)
      if reused_vid is not None and offset is not None:
        yield offset, reused_vid
      else:
        if offset is None:
          offset = 0
        if isinstance(schedule.left, CommSchedule):
          yield from get_attrs(schedule.left, reuses, offset)
        else:
          yield offset, 0
        offset += schedule.distance
        if isinstance(schedule.right, CommSchedule):
          yield from get_attrs(schedule.right, reuses, offset)
        else:
          yield offset, 0

    cr_vars = OrderedDict([(self, 1)])
    cr_vars_table = {1: self}
    # var_0 is input, var_1 is output
    for child, count in OrderedCounter(self.children).items():
      if count > 1:
        cr_vars[child] = len(cr_vars) + 1
        cr_vars_table[len(cr_vars)] = child
    for schedule, vid in cr_vars.items():
      _logger.debug('var_%d: %s', vid, schedule)

    vars_to_process = collections.deque([self])
    vars_processed = {0}
    dependers = OrderedDict()  # type: Dict[int, Dict[int, None]]
    dependees = OrderedDict()  # type: Dict[int, Dict[int, Tuple[int, int]]]
    while vars_to_process:
      schedule = vars_to_process.popleft()
      dst_vid = cr_vars[schedule]
      vars_processed.add(dst_vid)
      for offset, src_vid in get_attrs(schedule, cr_vars):
        _logger.debug('var_%d accesses var_%d @ %d', dst_vid, src_vid, offset)
        dependers.setdefault(src_vid, OrderedDict()).setdefault(dst_vid, None)
        dependees.setdefault(dst_vid,
                             OrderedDict()).setdefault(src_vid,
                                                       (offset, offset))
        min_offset, max_offset = dependees[dst_vid][src_vid]
        dependees[dst_vid][src_vid] = (min(offset,
                                           min_offset), max(offset, max_offset))
        if (src_vid not in vars_processed and
            cr_vars_table[src_vid] not in vars_to_process):
          vars_to_process.append(cr_vars_table[src_vid])

    def inline() -> Iterator[Tuple[int, int]]:
      """Inline a variable if it is only accessed once by another variable.

      Yields:
        Tuple of variable ids of inlined source and destination variables.
      """
      if _logger.isEnabledFor(logging.DEBUG):
        _logger.debug('looking for inline')
        _logger.debug('  dependers:')
        for src_vid, dst_vids in dependers.items():
          _logger.debug('    var_%d is accessed by %s', src_vid,
                        ', '.join(map('var_{}'.format, dst_vids)))
        _logger.debug('  dependees:')
        for dst_vid in dependees:
          for src_vid, (min_offset, max_offset) in dependees[dst_vid].items():
            _logger.debug('    var_%d accesses var_%d @ [%d, %d]', dst_vid,
                          src_vid, min_offset, max_offset)

      for src_vid, dst_vids in dependers.items():
        if len(dst_vids) == 1:
          dst_vid = tuple(dst_vids)[0]
          min_offset, max_offset = dependees[dst_vid][src_vid]
          if min_offset == max_offset:
            _logger.debug(
                'var_%d is only accessed by var_%d @ %d, '
                'it should be inlined', src_vid, dst_vid, min_offset)
            yield src_vid, dst_vid
            yield from inline()
            break
      else:
        _logger.debug('  cannot find inline oppotunities')

    for src_vid, dst_vid in inline():
      _logger.debug('inlining var_%d', src_vid)
      offset = dependees[dst_vid][src_vid][0]
      for src_src_vid, (min_offset, max_offset) in dependees[src_vid].items():
        new_min_offset = min_offset + offset
        new_max_offset = max_offset + offset
        _logger.debug('var_%d accesses var_%d @ %d', dst_vid, src_vid, offset)
        _logger.debug('var_%d accesses var_%d @ [%d, %d]', src_vid, src_src_vid,
                      min_offset, max_offset)
        _logger.debug('therefore var_%d accesses var_%d @ [%d, %d] via var_%d',
                      dst_vid, src_src_vid, new_min_offset, new_max_offset,
                      src_vid)
        if src_src_vid in dependees[dst_vid]:
          _logger.debug('var_%d used to access var_%d @ [%d, %d]', dst_vid,
                        src_src_vid, *dependees[dst_vid][src_src_vid])
        old_min_offset, old_max_offset = dependees[dst_vid].get(
            src_src_vid, (new_min_offset, new_max_offset))
        dependees[dst_vid][src_src_vid] = (min(old_min_offset, new_min_offset),
                                           max(old_max_offset, new_max_offset))
        _logger.debug('after inlining, var_%d accesses var_%d @ [%d, %d]',
                      dst_vid, src_src_vid, *dependees[dst_vid][src_src_vid])
      src_src_vids = list(dependees[src_vid])
      for src_src_vid in src_src_vids:
        dependers[src_src_vid][dst_vid] = None
        del dependers[src_src_vid][src_vid]
      del dependers[src_vid]
      del dependees[dst_vid][src_vid]
      del dependees[src_vid]
      del cr_vars_table[src_vid]
    self._dependers, self._dependees = dependers, dependees
    self._cr_vars_table = cr_vars_table

  @property
  def dependers(self) -> Dict[int, Dict[int, None]]:
    """The dependers table.

    Returns:
      A dict mapping variable ids to its dependers set, which is an ordered dict
      mapping dependers to None.
    """
    if not hasattr(self, '_dependers'):
      self._calc_dependency()
    return self._dependers

  @property
  def dependees(self) -> Dict[int, Dict[int, Tuple[int, int]]]:
    """The dependees table.

    Returns:
      A dict mapping variable ids to its dependees dict, which maps dependees to
      a tuple of the first and last accessed offset.
    """
    if not hasattr(self, '_dependees'):
      self._calc_dependency()
    return self._dependees

  @property
  def cr_vars_table(self) -> Dict[int, 'CommSchedule']:
    if not hasattr(self, '_cr_vars_table'):
      self._calc_dependency()
    return self._cr_vars_table

  @cached_property.cached_property
  def total_distance(self) -> int:
    """Calculate the total reuse distance.

    The total reuse distance is a measure of hardware resource cost in terms of
    required reuse buffer size. It is calculated by summing up the reuse
    distance of all reused variables. The reused distance is determined by the
    offset when a variable is first produced and last consumed. The latter is
    static and the former is calculated via solving an ILP.

    Returns:
      The total reuse distance.
    """

    # solve ILP for optimal offsets
    lp_problem = pulp.LpProblem("optimal_offsets", pulp.LpMinimize)
    lp_vars = {0: 0, 1: 0}
    lp_helper_vars = {}
    objectives = []
    for src_vid in self.dependers:
      lp_var = pulp.LpVariable('produced_offset_%d' % src_vid, cat='Integer')
      lp_helper_var = pulp.LpVariable('consumed_offset_%d' % src_vid,
                                      cat='Integer')
      lp_vars.setdefault(src_vid, lp_var)
      lp_helper_vars[src_vid] = lp_helper_var
      objectives.append(lp_helper_var - lp_vars[src_vid])
    lp_problem += sum(objectives)
    for src_vid, dst_vids in self.dependers.items():
      for dst_vid in dst_vids:
        min_offset, max_offset = self.dependees[dst_vid][src_vid]
        lp_problem += lp_vars[src_vid] <= min_offset + lp_vars[dst_vid]
        lp_problem += lp_helper_vars[src_vid] >= max_offset + lp_vars[dst_vid]
    lp_status = pulp.LpStatus[lp_problem.solve()]
    _logger.debug('ILP status: %s', lp_status)
    _logger.debug('var_0 should be produced @ 0 and kept until %d',
                  int(pulp.value(lp_helper_vars[0])))
    self._produce_offsets = {}  # type: Dict[int, int]
    self._consume_offsets = {}  # type: Dict[int, int]
    for vid, lp_var in lp_vars.items():
      if vid == 1:  # output
        continue
      produce_offset = 0 if vid == 0 else int(pulp.value(lp_var))
      consume_offset = int(pulp.value(lp_helper_vars[vid]))
      _logger.debug('var_%d should be produced @ %d and kept until %d', vid,
                    produce_offset, consume_offset)
      self._produce_offsets[vid] = produce_offset
      self._consume_offsets[vid] = consume_offset
    total_distance = int(pulp.value(lp_problem.objective))
    max_rattr = max(
        attr if isinstance(attr, int) else attr[0] for attr in self.norm_attrs)
    assert max_rattr <= total_distance, '%s < %s' % (total_distance, max_rattr)
    return total_distance

  @property
  def produce_offsets(self) -> Dict[int, int]:
    """Maps variable ids to offsets at which they are first produced"""
    if not hasattr(self, '_produce_offsets'):
      _ = self.total_distance
    return self._produce_offsets

  @property
  def consume_offsets(self) -> Dict[int, int]:
    """Maps variable ids to offsets at which they are last consumed"""
    if not hasattr(self, '_consume_offsets'):
      _ = self.total_distance
    return self._consume_offsets

  def get_attrs_with_offset(self, offset: int = 0) -> Iterator[Attr]:
    """Generate all attributes with the given offset.

    Args:
      offset: The offset of the smallest relative attribute.

    Yields:
      Attributes in this schedule, NOT necessarily sorted by their relative
      attributes.
    """
    if isinstance(self.left, CommSchedule):  # Left child is not a leaf.
      yield from self.left.get_attrs_with_offset(offset)
    else:  # Left child is a leaf.
      if self.aattrs is None:  # Null absolute attributes.
        yield offset
      else:  # Non-null absolute attributes.
        yield offset, self.left

    offset += self.distance

    if isinstance(self.right, CommSchedule):  # Right child is not a leaf.
      yield from self.right.get_attrs_with_offset(offset)
    else:  # Right child is a leaf.
      if self.aattrs is None:  # Null absolute attributes.
        yield offset
      else:  # Non-null absolute attributes.
        yield offset, self.right

  @property
  def norm_attrs(self) -> Iterator[Attr]:
    return self.get_attrs_with_offset()

  @cached_property.cached_property
  def norm_attr_set(self) -> FrozenSet[Attr]:
    return frozenset(self.norm_attrs)

  @cached_property.cached_property
  def uniq_expr_set(self) -> Set[FrozenSet[Attr]]:
    """Unique expressions of this schedule.

    Returns:
      A dict mapping norm_attr_sets to a list of schedules whose normalized
      attributes equals the keys.
    """
    exprs = set()
    exprs.add(self.norm_attr_set)
    for child in self.left, self.right:
      if isinstance(child, CommSchedule):
        exprs |= child.uniq_expr_set
    return exprs

  @property
  def uniq_expr_dict(self) -> Dict[FrozenSet[Attr], List['CommSchedule']]:
    """Unique expressions of this schedule.

    Returns:
      A dict mapping norm_attr_sets to a list of schedules whose normalized
      attributes equals the keys.
    """
    exprs = OrderedDict()  # type: Dict[FrozenSet[Attr], List[CommSchedule]]
    exprs[self.norm_attr_set] = [self]
    for child in self.left, self.right:
      if isinstance(child, CommSchedule):
        for attrs, schedules in child.uniq_expr_dict.items():
          exprs.setdefault(attrs, []).extend(schedules)
    return exprs

  def get_ir_node_with_offset(self, offset: int = 0) -> ir.Node:
    """Get the IR node with the given offset.

    Args:
      offset: The offset of the smallest relative attribute.

    Returns:
      An ir.BinaryOp as the root of the IR.
    """
    if isinstance(self.left, CommSchedule):  # Left child is not a leaf.
      left_child = self.left.get_ir_node_with_offset(offset)  # type: ir.Node
    else:  # Left child is a leaf.
      left_child = assemble_attr(self.linearizer(offset),
                                 self.aattr_table[self.left])

    offset += self.distance

    if isinstance(self.right, CommSchedule):  # Right child is not a leaf.
      right_child = self.right.get_ir_node_with_offset(offset)  # type: ir.Node
    else:  # Right child is a leaf.
      right_child = assemble_attr(self.linearizer(offset),
                                  self.aattr_table[self.right])

    return ir.from_reduction(self.operator, (left_child, right_child))

  @cached_property.cached_property
  def ir_node(self) -> ir.Node:
    return self.get_ir_node_with_offset(self.rattrs[0])

  @cached_property.cached_property
  def _rcr_write_idx_table(self) -> Dict[ir.BinaryOp, Tuple[int, ...]]:
    """Returns the CSE write index table.

    Returns:
      A dict mapping normalized common subexpressions to the index at which the
      new variable should write to.
    """
    table = {}
    for vid in self.dependers:
      if vid == 0:  # 0 is the input
        continue
      expr = mutator.normalize(self.cr_vars_table[vid].ir_node)
      # okay not to add references because expr does not contain cr vars
      # TODO: support multi-stage input
      table[mutator.normalize(expr)] = util.add_inv(
          soda.visitor.get_normalize_index(expr))
    return table

  def get_ir_node_with_rcr(self,
                           stencil,
                           rcrs: MutableMapping[ir.BinaryOp, ir.Ref],
                           write_idx_table=None) -> ir.Node:
    """Returns an ir.Node with relative computation reuse.

    Args:
      stencil: soda.core.Stencil object, whose symbol table will be updated.
      rcrs: Dict mapping common subexpressions to the ir.Ref representing the
        write access to the new variable. This dict is used as both an input and
        an output. If the keys contain other common subexpressions, they should
        be replaced with the new variables.
      cse_write_idx_table: The CSE write index table, used to detect common
        subexpressions.
    """
    if write_idx_table is None:
      write_idx_table = self._rcr_write_idx_table
    operands = []  # type: List[ir.Node]
    _logger.debug('get ir node with cse for %s', self)
    if _logger.isEnabledFor(logging.DEBUG):
      for rattr, aattr in self:
        _logger.debug('  rattr: %s aattr: %s', rattr, aattr)
    for rattr, aattr in (  # type: ignore
        (self.rattrs[0], self.left), (self.rattrs[0] + self.distance,
                                      self.right)):
      # replace the aattr with proper normalized node
      if isinstance(aattr, CommSchedule):
        # node_without_cse is used for key-ing write_idx_table
        # okay not to add references because expr does not contain cr vars
        # TODO: support multi-stage input
        node_without_cse = mutator.shift(
            aattr.ir_node, soda.visitor.get_normalize_index(aattr.ir_node))
        node_with_cse = aattr.get_ir_node_with_rcr(stencil, rcrs,
                                                   write_idx_table)
        # node_with_cse_norm is used for key-ing rcrs
        node_with_cse_norm = mutator.normalize(
            node_with_cse, {ref.name: ref.idx for ref in rcrs.values()})
        idx = write_idx_table.get(node_without_cse)
        if idx is not None:
          # this aattr is a common subexpression, need to be replaced
          # cses is key-ed by expressions with recursive CSE applied
          if node_with_cse_norm not in rcrs:
            # this is the first time we see this aattr, allocate a new name
            node = ir.Ref(name=stencil.new_cr_var(), idx=idx, lat=None)
            _logger.debug('  replace %s with %s', node_without_cse, node)
            stencil.symbol_table[node.name] = node_without_cse.haoda_type
            rcrs[node_with_cse_norm] = node
          else:
            # we have seen this aattr, replace with the known
            node = rcrs[node_with_cse_norm]
        else:
          # this aattr is not a common subexpression
          node = mutator.shift(node_with_cse, self.linearizer(rattr))
      else:
        node = self.aattr_table[aattr]
      operands.append(assemble_attr(self.linearizer(rattr), node))

    return arithmetic.simplify(ir.from_reduction(self.operator,
                                                 tuple(operands)))

  def get_ir_node_with_cr(self, stencil, crs: Dict[ir.BinaryOp,
                                                   ir.Ref]) -> ir.Node:
    rcrs = crs.copy()
    ir_node_with_rcr = self.get_ir_node_with_rcr(stencil, rcrs)
    norm_refs = {getattr(ref, 'name'): ref.idx for ref in rcrs.values()}

    # try to apply absolute CSE
    binary_aattrs = collections.defaultdict(
        list)  # type: Dict[ir.BinaryOp, List[Tuple[int, ...]]]

    def add_to_count(node: ir.Node, norm_idx: Index = ()) -> None:
      reduction = ir.to_reduction(node)
      if reduction is not None:
        for op in reduction[1]:
          if isinstance(op, ir.BinaryOp):
            idx = soda.visitor.get_normalize_index(op, references=norm_refs)
            if norm_idx:
              idx = tuple(x - y for x, y in zip(idx, norm_idx))
            binary_aattrs[mutator.normalize(op,
                                            references=norm_refs)].append(idx)

    _logger.debug('counting complex aattrs')
    norm_idx = soda.visitor.get_normalize_index(self.ir_node,
                                                references=norm_refs)
    add_to_count(ir_node_with_rcr, norm_idx)  # not normalized; need norm_idx
    for tcs in rcrs:
      add_to_count(tcs)  # all normalized
    _logger.debug('complex aattrs count')
    acrs = {}
    for op, indices in binary_aattrs.items():
      count = len(indices)
      _logger.debug(' %s x%d', op, count)
      if count > 1:
        new_name = stencil.new_cr_var()
        # select the least index to minimize buffer size required
        min_idx = min(indices, key=lambda x: tuple(reversed(x)))
        acrs[op] = ir.Ref(name=new_name, idx=util.add_inv(min_idx), lat=None)
        stencil.symbol_table[new_name] = getattr(op, 'haoda_type')

    do_acr = lambda op: mutator.replace_expressions(
        op, acrs, references=norm_refs)
    rcrs = OrderedDict((do_acr(k), v) for k, v in rcrs.items())
    crs.update(rcrs)
    crs.update(acrs)
    if _logger.isEnabledFor(logging.INFO):
      for tcs, ref in rcrs.items():
        _logger.info('rcr: %s => %s', ir.unparenthesize(tcs), ref)
      for tcs, ref in acrs.items():
        _logger.info('acr: %s => %s', ir.unparenthesize(tcs), ref)

    reduction = ir.to_reduction(ir_node_with_rcr)
    assert reduction is not None
    return arithmetic.simplify(
        ir.from_reduction(reduction[0], tuple(map(do_acr, reduction[1]))))


def make_schedule_from_json(j: Dict[str, Any], offset: int,
                            null_aattr: bool) -> CommSchedule:
  left, right, distance = j['left'], j['right'], j['distance']
  attrs = []  # type: List[Tuple[RelativeAttr, AbsoluteAttr]]
  if isinstance(left, dict):
    left = make_schedule_from_json(left, offset, null_aattr)
    attrs.extend(left)
  elif isinstance(left, int):
    if null_aattr:
      left = None
    attrs.append((offset, left))

  offset += distance

  if isinstance(right, dict):
    right = make_schedule_from_json(right, offset, null_aattr)
    attrs.extend(right)
  elif isinstance(right, int):
    if null_aattr:
      right = None
    attrs.append((offset, right))

  attrs.sort(key=lambda attr: attr[0])
  rattrs, aattrs = zip(*attrs)
  return CommSchedule(left, right, distance, rattrs,
                      None if null_aattr else aattrs)


class CommSchedules(ScheduleBase):
  """Schedules of an Expression.

  Class Attributes:
    range_func: The range function to use for range(n).
    skip: Whether to skip iterations if cost has exceeded the current minimum.
    lazy: Whether to evaluate the Cartesian product lazily.

  Attributes:
    operands: String of binary mask of the operands.
    cache: A mapping from operands to CommSchedules, or None.
    stat: A list of [cache_hit, cache_miss, loop 1 trip count, loop 2 trip
        count, loop 3 trip count], or None.
    max_cost: The cut-off cost, or None. If not None, any schedule must has
        less cost than this value to be included in the schedules.
  """
  range_func = range_from_middle
  skip = True
  lazy = True

  @staticmethod
  def set_optimizations(optimizations: Sequence[str]) -> None:
    if 'reorder-exploration' in optimizations:
      CommSchedules.range_func = range_from_middle
    if 'no-reorder-exploration' in optimizations:
      CommSchedules.range_func = lambda n: iter(range(n))
    if 'skip-with-partial-cost' in optimizations:
      CommSchedules.skip = True
    if 'no-skip-with-partial-cost' in optimizations:
      CommSchedules.skip = False
    if 'lazy-evaluation' in optimizations:
      CommSchedules.lazy = True
    if 'no-lazy-evaluation' in optimizations:
      CommSchedules.lazy = False

  def __init__(
      self,
      rattrs: Tuple[RelativeAttr, ...],
      aattrs: Optional[Tuple[AbsoluteAttr, ...]] = None,
      operands: Optional[str] = None,
      # pylint: disable=unsubscriptable-object
      cache: Optional[Dict[Tuple[int, ...], 'CommSchedules']] = None,
      stat: Optional[List[int]] = None,
      max_cost: Optional[int] = None,
      timeout: Optional[int] = None) -> None:
    super().__init__(rattrs, aattrs)
    if operands is None:
      self.operands = '1' * len(self.rattrs)
    else:
      self.operands = operands
    self.cache = cache
    if cache is not None:
      cache[self.key(self.operands)] = self
    if stat is None:
      self.stat = [0, 0, 0, 0, 0]
    else:
      self.stat = stat
    if max_cost is None:
      self.max_cost = collections.Counter(self.operands)['1']
    else:
      self.max_cost = max_cost
    self.timeout = 300
    if timeout is not None:
      self.timeout = timeout

  def __iter__(self) -> Iterator[CommSchedule]:  # type: ignore
    if hasattr(self, 'schedules'):
      return iter(getattr(self, 'schedules'))
    return self.generator

  def key(self, operands) -> Tuple[int, ...]:
    offset = self.rattrs[next(
        idx for idx, bit in enumerate(operands) if bit == '1')]
    key = [
        self.rattrs[idx] - offset
        for idx, bit in enumerate(operands)
        if bit == '1'
    ]
    if self.aattrs is not None:
      key.extend(
          self.aattrs[idx] for idx, bit in enumerate(operands) if bit == '1')
    return tuple(key)

  @property
  def generator(self) -> Iterator[CommSchedule]:
    """Generates possible schedules via dynamic programming.

    This generator will lazily materialize sub-problems for the Cartesian
    product.

    Yields:
      One CommSchedule at a time. If self.skip is on, the cost of generated
      CommSchedule with be monotonically decreasing.
    """
    n = collections.Counter(self.operands)['1']
    num_operands = len(self.rattrs)
    indices = [i for i in range(num_operands) if self.operands[i] == '1']
    schedules = []
    skipped = False
    if n == 1:
      schedule = self.aattrs[indices[0]] if self.aattrs is not None else None
      schedules.append(schedule)
      self.schedules = schedules
      self.max_cost = 0
      yield schedule  # type: ignore
      return
    #if self.aattrs is None or len(set(
    #    self.aattrs[i] for i in range(num_operands)
    #    if self.operands[i] == '1')) == 1:
    #  selector = lambda indices, m: [tuple(indices[1:m])]
    #  _logger.debug('using associative selector for %s', self.operands)
    #else:
    #  selector = lambda indices, m: itertools.combinations(indices[1:], m)
    #  _logger.debug('using commutative selector for %s', self.operands)
    selector = lambda indices, m: itertools.combinations(indices[1:], m)
    for m in CommSchedules.range_func(n - 1):
      selections = selector(indices, m)
      for selection in selections:
        self.stat[2] += 1
        left_indices = (indices[0],) + selection
        right_indices = [i for i in indices if i not in left_indices]
        left_operands = ''.join(
            '1' if i in left_indices else '0' for i in range(num_operands))
        right_operands = ''.join(
            '1' if i in right_indices else '0' for i in range(num_operands))
        for left in self.get_schedules(left_operands):
          self.stat[3] += 1
          left_cost = 1
          if isinstance(left, CommSchedule):
            left_cost += left.num_ops
          if self.skip and left_cost > self.max_cost:
            skipped = True
            continue
          for right in self.get_schedules(right_operands):
            self.stat[4] += 1
            right_cost = 1
            if isinstance(right, CommSchedule):
              right_cost += right.num_ops
            if self.skip and right_cost > self.max_cost:
              skipped = True
              continue
            distance = self.rattrs[right_indices[0]]
            distance -= self.rattrs[left_indices[0]]
            rattrs = tuple(self.rattrs[i]
                           for i, op in enumerate(self.operands)
                           if op != '0')
            aattrs = None
            if self.aattrs is not None:
              aattrs = tuple(self.aattrs[i]
                             for i, op in enumerate(self.operands)
                             if op != '0')
            schedule = CommSchedule(  # type: ignore
                left, right, distance, rattrs, aattrs)
            num_ops = schedule.num_ops  # type: ignore
            if num_ops < self.max_cost:
              self.max_cost = num_ops
            schedules.append(schedule)
            yield schedule  # type: ignore
    self.schedules = schedules

  schedule_cache = {}  # type: Dict[CommSchedule, CommSchedule]

  def make_schedule(self, left, right, distance):
    new_schedule = CommSchedule(left, right, distance, self.rattrs, self.aattrs)
    schedule = CommSchedules.schedule_cache.get(new_schedule)
    if schedule is not None:
      return schedule
    CommSchedules.schedule_cache[new_schedule] = new_schedule
    return new_schedule

  @property
  def best(self) -> CommSchedule:
    best = None
    try:
      with util.timeout(self.timeout):
        for schedule in self:
          if best is None or schedule.cost < best.cost:
            best = schedule
            _logger.debug('schedule: %s', best)
            _logger.info('cost: %s', best.cost)
    except TimeoutError:
      pass
    self.print_stats()
    if best is None:
      raise util.InternalError('cannot find best schedule')
    return best

  @property
  def cache_hit(self) -> int:
    return self.stat[0]

  @property
  def cache_miss(self) -> int:
    return self.stat[1]

  @property
  def cache_hit_rate(self) -> float:
    try:
      return self.cache_hit / (self.cache_hit + self.cache_miss)
    except ZeroDivisionError:
      return float('nan')

  def get_schedules(self, operands: str) -> Iterator[CommSchedule]:
    """Get schedules with the given operands.

    If self.cache is not None and the same arguments were given in previous runs
    of this object, the result will be fetched from the cache.

    Args:
      operands: Bit-mask of the operands.
    Returns:
      CommSchedules with the given operands.
    """
    if self.cache is not None:
      schedules = self.cache.get(self.key(operands))
      if schedules is not None:
        self.stat[0] += 1
        if hasattr(schedules, 'schedules'):
          return iter(schedules.schedules)  # type: ignore
        return schedules.generator
    self.stat[1] += 1
    return CommSchedules(self.rattrs,
                         self.aattrs,
                         operands=operands,
                         cache=self.cache,
                         stat=self.stat,
                         max_cost=min(
                             self.max_cost,
                             collections.Counter(operands)['1'])).generator

  def print_stats(self, logger: Callable[..., None] = _logger.info) -> None:
    logger('loops: | L1: %d | L2: %d | L3: %d |', *self.stat[2:])
    logger('cache: | hit: %d | miss: %d | hit rate: %2.3f %% |', self.cache_hit,
           self.cache_miss, self.cache_hit_rate * 100)


class GreedySchedules(ScheduleBase):
  """Schedules of an Expression, found greedily.
  """
  timeout = 1
  num_pruned = 5

  def __init__(self,
               rattrs: Tuple[RelativeAttr, ...],
               aattrs: Optional[Tuple[Union[AbsoluteAttr, CommSchedule, None],
                                      ...]] = None,
               linearizer: Optional[Linearizer] = None,
               cache: Optional[Dict] = None) -> None:
    self.linearizer = linearizer
    super().__init__(rattrs, aattrs)  # type: ignore

  def __lt__(self, other: 'GreedySchedules') -> bool:
    return self.comparison_key.cost < other.comparison_key.cost

  @cached_property.cached_property
  def comparison_key(self) -> CommSchedule:
    return linear_schedule(tuple(self))

  @property
  def generator(self) -> Iterator[CommSchedule]:
    attr_map = {attr: idx for idx, attr in enumerate(self)}
    reuses = OrderedDict()  # type: Dict[CommSchedule, List[Tuple[int, int]]]
    has_conflict = collections.defaultdict(
        lambda: False)  # type: Dict[CommSchedule, bool]
    for left, right in itertools.combinations(self, 2):
      left_rattr, left_aattr = left
      right_rattr, right_aattr = right
      distance = right_rattr - left_rattr
      new_aattr = left_aattr, right_aattr
      operation = CommSchedule(left_aattr, right_aattr, distance,
                               (left_rattr, right_rattr),
                               new_aattr)  # type: ignore
      if operation in reuses:
        continue
      #_logger.debug('look for operation: %s', operation)
      # look for reuse of this operation over all operands
      reuses[operation] = []
      # conflict groups
      # group_lists maps group id to a list of (idx_l, idx_r) tuple, in order
      # group_table maps attr idx to group id
      group_lists = []  # type: List[List[Tuple[int, int]]]
      group_table = {}  # type: Dict[int, int]
      for idx_l, (rattr_l, aattr_l) in enumerate(self):
        if aattr_l != left_aattr:
          continue
        rattr_r, aattr_r = rattr_l + right_rattr - left_rattr, right_aattr
        idx_r = attr_map.get((rattr_r, aattr_r))
        if idx_r is None:
          continue
        # (idx_l, idx_r) is compatible with `operation`
        group_id = group_table.get(idx_l)
        if group_id is None:
          group_id = group_table.get(idx_r)
        if group_id is None:
          # new conflict group
          group_id = len(group_lists)
          group_lists.append([])
        group_lists[group_id].append((idx_l, idx_r))
        group_table[idx_l] = group_id
        group_table[idx_r] = group_id

      for group_list in group_lists:
        if len(group_list) > 1:
          _logger.debug('conflict group of %s: %s', operation, group_list)
          has_conflict[operation] = True

      for group_list in group_lists:
        if len(group_list) % 2 != 0:
          reuses[operation].extend(group_list[::2])

      min_idx_l = min((x[0] for x in reuses[operation]), default=0)
      max_idx_l = max((x[0] for x in reuses[operation]), default=-1)
      _logger.debug('min_idx_l: %s | max_idx_l: %s', min_idx_l, max_idx_l)

      for group_list in group_lists:
        if len(group_list) % 2 == 0:
          span_0 = (self.rattrs[max(group_list[-2][0], max_idx_l)] -
                    self.rattrs[min(group_list[0][0], min_idx_l)])
          span_1 = (self.rattrs[max(group_list[-1][0], max_idx_l)] -
                    self.rattrs[min(group_list[1][0], min_idx_l)])
          _logger.debug('span 0: %d, span 1: %d', span_0, span_1)
          reuses[operation].extend(group_list[1 if span_1 < span_0 else 0::2])
      if len(reuses[operation]) > 1:
        _logger.debug('reuses[%s]: %s', operation, reuses[operation])
      reuses[operation].sort()  # ?

    # filter out operations that cannot be reused
    # they may not all be useful because they overlap
    reuses = {k: v for k, v in reuses.items() if len(v) > 1}
    if not reuses:
      yield linear_schedule(tuple(self))
      return

    def aligns(dis: int, dim: int) -> bool:
      """Returns whether dis aligns with dim.

      A distance aligns with a dimension if the indices of two points with that
      distance differ and only differ in that dimension.
      """
      assert self.linearizer is not None
      zipped = zip(self.linearizer(dis), self.linearizer.mins,
                   self.linearizer.dims)
      return all(idx != min_idx if d == dim else idx == min_idx
                 for idx, min_idx, d in zipped)

    if self.linearizer is not None and len(reuses) > len(self):
      _logger.debug('linearizer: mins: %s maxs: %s', self.linearizer.mins,
                    self.linearizer.maxs)
      for dim in reversed(self.linearizer.dims):
        if any(aligns(op.distance, dim) for op in reuses):
          # only consider reuse with a single dimension, if possible
          reuses = {
              k: [(idx_l, idx_r)
                  for idx_l, idx_r in v
                  if aligns(self.rattrs[idx_r] - self.rattrs[idx_l], dim)
                 ] for k, v in reuses.items() if aligns(k.distance, dim)
          }
          break

    candidates = []
    for op in reuses:
      # find all compatible reuses that include op
      _logger.debug('find all compatible reuses that include %s', op)
      new_attrs = OrderedDict(enumerate(self))
      used = set()  # type: Set[int]

      def do_reuse_for(schedule: CommSchedule) -> None:
        reused_indices = [(idx_l, idx_r)
                          for idx_l, idx_r in reuses[schedule]
                          if idx_l not in used and idx_r not in used]
        if len(reused_indices) > 1:
          for idx_l, idx_r in reused_indices:
            _logger.debug('reusing %s for %s + %s', schedule,
                          '%s:%s' % self[idx_l], '%s:%s' % self[idx_r])
            # pylint: disable=cell-var-from-loop
            new_attrs[idx_l] = new_attrs[idx_l][0], schedule  # type: ignore
            del new_attrs[idx_r]
            used.update({idx_l, idx_r})

      do_reuse_for(op)
      for operation in sorted(reuses,
                              key=lambda s: (-len(reuses[s]), s.distance)):
        do_reuse_for(operation)

      new_rattrs, new_aattrs = zip(*new_attrs.values())
      candidates.append((has_conflict[op],
                         GreedySchedules(new_rattrs, new_aattrs,
                                         self.linearizer)))

    nsmallest = heapq.nsmallest(GreedySchedules.num_pruned, candidates)

    if _logger.isEnabledFor(logging.DEBUG):
      for conflict, schedule in nsmallest:
        _logger.debug('candidate: %s cost: %s', schedule.comparison_key,
                      (conflict, *schedule.comparison_key.cost))

    for _, schedule in nsmallest:
      yield from schedule.generator

  @cached_property.cached_property
  def best(self) -> CommSchedule:
    generator = self.generator
    best = next(generator)
    try:
      with util.timeout(GreedySchedules.timeout):
        for schedule in generator:
          _logger.debug('schedule: %s', schedule)
          if schedule.cost < best.cost:
            best = schedule
            _logger.info('schedule: %s num ops: %d total distance: %d', best,
                         *best.cost)
    except TimeoutError:
      _logger.warning('timeout after %s sec', self.timeout)
    self.print_stats()
    _logger.info('schedule: %s num ops: %d total distance: %d', best,
                 *best.cost)
    return best

  def print_stats(self, logger: Callable[..., None] = _logger.info) -> None:
    return


class BeamSchedules(ScheduleBase):
  """Schedules of an Expression, found using beam search.
  """
  beam_width = 5

  def __init__(self,
               rattrs: Tuple[RelativeAttr, ...],
               aattrs: Optional[Tuple[Union[AbsoluteAttr, CommSchedule, None],
                                      ...]] = None,
               linearizer: Optional[Linearizer] = None,
               cache: Optional[Dict] = None) -> None:
    self.linearizer = linearizer
    super().__init__(rattrs, aattrs)  # type: ignore

  def __lt__(self, other: 'BeamSchedules') -> bool:
    return self.linear_schedule.cost < other.linear_schedule.cost

  @cached_property.cached_property
  def linear_schedule(self) -> CommSchedule:
    return linear_schedule(tuple(self))

  @property
  def candidates(self) -> List[Tuple[bool, 'BeamSchedules']]:
    """Return new candidates with computation reuse.

    Returns:
        List[Tuple[bool, BeamSchedules]]: A list of (has_conflict, candidates).
    """

    attr_map = {attr: idx for idx, attr in enumerate(self)}
    reuses = OrderedDict()  # type: Dict[CommSchedule, List[Tuple[int, int]]]
    has_conflict = collections.defaultdict(
        lambda: False)  # type: Dict[CommSchedule, bool]
    for left, right in itertools.combinations(self, 2):
      left_rattr, left_aattr = left
      right_rattr, right_aattr = right
      distance = right_rattr - left_rattr
      new_aattr = left_aattr, right_aattr
      operation = CommSchedule(left_aattr, right_aattr, distance,
                               (left_rattr, right_rattr),
                               new_aattr)  # type: ignore
      if operation in reuses:
        continue
      #_logger.debug('look for operation: %s', operation)
      # look for reuse of this operation over all operands
      reuses[operation] = []
      # conflict groups
      # group_lists maps group id to a list of (idx_l, idx_r) tuple, in order
      # group_table maps attr idx to group id
      group_lists = []  # type: List[List[Tuple[int, int]]]
      group_table = {}  # type: Dict[int, int]
      for idx_l, (rattr_l, aattr_l) in enumerate(self):
        if aattr_l != left_aattr:
          continue
        rattr_r, aattr_r = rattr_l + right_rattr - left_rattr, right_aattr
        idx_r = attr_map.get((rattr_r, aattr_r))
        if idx_r is None:
          continue
        # (idx_l, idx_r) is compatible with `operation`
        group_id = group_table.get(idx_l)
        if group_id is None:
          group_id = group_table.get(idx_r)
        if group_id is None:
          # new conflict group
          group_id = len(group_lists)
          group_lists.append([])
        group_lists[group_id].append((idx_l, idx_r))
        group_table[idx_l] = group_id
        group_table[idx_r] = group_id

      for group_list in group_lists:
        if len(group_list) > 1:
          _logger.debug('conflict group of %s: %s', operation, group_list)
          has_conflict[operation] = True

      for group_list in group_lists:
        if len(group_list) % 2 != 0:
          reuses[operation].extend(group_list[::2])

      min_idx_l = min((x[0] for x in reuses[operation]), default=0)
      max_idx_l = max((x[0] for x in reuses[operation]), default=-1)
      _logger.debug('min_idx_l: %s | max_idx_l: %s', min_idx_l, max_idx_l)

      for group_list in group_lists:
        if len(group_list) % 2 == 0:
          span_0 = (self.rattrs[max(group_list[-2][0], max_idx_l)] -
                    self.rattrs[min(group_list[0][0], min_idx_l)])
          span_1 = (self.rattrs[max(group_list[-1][0], max_idx_l)] -
                    self.rattrs[min(group_list[1][0], min_idx_l)])
          _logger.debug('span 0: %d, span 1: %d', span_0, span_1)
          reuses[operation].extend(group_list[1 if span_1 < span_0 else 0::2])
      if len(reuses[operation]) > 1:
        _logger.debug('reuses[%s]: %s', operation, reuses[operation])
      reuses[operation].sort()  # ?

    # filter out operations that cannot be reused
    # they may not all be useful because they overlap
    reuses = {k: v for k, v in reuses.items() if len(v) > 1}
    if not reuses:
      return []

    def aligns(dis: int, dim: int) -> bool:
      """Returns whether dis aligns with dim.

      A distance aligns with a dimension if the indices of two points with that
      distance differ and only differ in that dimension.
      """
      assert self.linearizer is not None
      zipped = zip(self.linearizer(dis), self.linearizer.mins,
                   self.linearizer.dims)
      return all(idx != min_idx if d == dim else idx == min_idx
                 for idx, min_idx, d in zipped)

    if self.linearizer is not None and len(reuses) > len(self):
      _logger.debug('linearizer: mins: %s maxs: %s', self.linearizer.mins,
                    self.linearizer.maxs)
      for dim in reversed(self.linearizer.dims):
        if any(aligns(op.distance, dim) for op in reuses):
          # only consider reuse with a single dimension, if possible
          reuses = {
              k: [(idx_l, idx_r)
                  for idx_l, idx_r in v
                  if aligns(self.rattrs[idx_r] - self.rattrs[idx_l], dim)
                 ] for k, v in reuses.items() if aligns(k.distance, dim)
          }
          break

    candidates = []
    for op in reuses:
      # find all compatible reuses that include op
      _logger.debug('find all compatible reuses that include %s', op)
      new_attrs = OrderedDict(enumerate(self))
      used = set()  # type: Set[int]

      def do_reuse_for(schedule: CommSchedule) -> None:
        reused_indices = [(idx_l, idx_r)
                          for idx_l, idx_r in reuses[schedule]
                          if idx_l not in used and idx_r not in used]
        if len(reused_indices) > 1:
          for idx_l, idx_r in reused_indices:
            _logger.debug('reusing %s for %s + %s', schedule,
                          '%s:%s' % self[idx_l], '%s:%s' % self[idx_r])
            # pylint: disable=cell-var-from-loop
            new_attrs[idx_l] = new_attrs[idx_l][0], schedule  # type: ignore
            del new_attrs[idx_r]
            used.update({idx_l, idx_r})

      do_reuse_for(op)
      for operation in sorted(reuses,
                              key=lambda s: (-len(reuses[s]), s.distance)):
        do_reuse_for(operation)

      new_rattrs, new_aattrs = zip(*new_attrs.values())
      candidates.append((has_conflict[op],
                         BeamSchedules(new_rattrs, new_aattrs,
                                       self.linearizer)))

    return candidates

  @cached_property.cached_property
  def best(self) -> CommSchedule:
    candidates = [(False, self)]
    best = self.linear_schedule
    while candidates:
      candidates = heapq.nsmallest(
          BeamSchedules.beam_width,
          sum((x.candidates for _, x in candidates), []))
      for _, candidate in candidates:
        schedule = candidate.linear_schedule
        _logger.debug('schedule: %s', schedule)
        if schedule.cost < best.cost:
          best = schedule
          _logger.info('schedule: %s num ops: %d total distance: %d', best,
                       *best.cost)
    _logger.info('schedule: %s num ops: %d total distance: %d', best,
                 *best.cost)
    return best

  def print_stats(self, logger: Callable[..., None] = _logger.info) -> None:
    return


def linear_schedule(attrs) -> CommSchedule:
  """Schedule the attributes linearily.

  Args:
    indices: Iterable of ints, indicating what attributes should be used.

  Returns:
    CommSchedule of the attributes, scheduled as a linear tree.
  """
  rattrs, aattrs = zip(*attrs)
  # rattrs doesn't have to start with 0 but has to be sorted
  if list(rattrs) != sorted(rattrs):
    raise util.InputError('rattrs not sorted: %s' % str(rattrs))
  distance = rattrs[1] - rattrs[0]
  other_args = distance, rattrs, aattrs
  if len(attrs) == 2:
    return CommSchedule(aattrs[0], aattrs[1], *other_args)
  return CommSchedule(aattrs[0], linear_schedule(attrs[1:]), *other_args)


class GloreSchedules(ScheduleBase):
  """Schedules of an Expression, found using GLORE paper's heuristics.
  """

  def __init__(self,
               rattrs: Tuple[RelativeAttr, ...],
               aattrs: Optional[Tuple[Union[AbsoluteAttr, CommSchedule, None],
                                      ...]] = None,
               linearizer: Optional[Linearizer] = None,
               cache: Optional[Dict] = None) -> None:
    if linearizer is None:
      raise util.InputError('linearizer must not be None for GloreSchedules')
    self.linearizer = linearizer
    super().__init__(rattrs, aattrs)  # type: ignore

  @property
  def generator(self) -> Iterator[CommSchedule]:
    num_dim = self.linearizer.num_dim
    for direction in ((1,) + (0,) * (num_dim - 1), (1,) * num_dim):
      _logger.debug('explore direction %s', direction)

      # maps line_id to (attrs, reuse distance list)
      # line_id is the unique identifier of a line along the specific direction
      reuse_groups = collections.defaultdict(lambda: ([], [
      ]))  # type: Dict[Index, Tuple[List[Tuple[Index, Any]], List[int]]]

      def log_reuse_group(logger=_logger.debug,
                          reuse_groups=reuse_groups,
                          direction=direction) -> None:
        for line_id, (group, reuse_distance_list) in reuse_groups.items():
          logger('logging reuse groups')
          logger('  line #%s along direction <%s>:', line_id, direction)
          rattr_indices, aattrs = zip(*group)
          logger('    rattrs: %s', util.lst2str(rattr_indices))
          logger('    aattrs: %s', util.lst2str(aattrs))
          logger('    reuse distances: %s', reuse_distance_list)
        logger('')

      # step 1: grouping
      _logger.debug('attrs: %s', util.lst2str(self))
      for rattr, aattr in self:
        rattr_idx = self.linearizer(rattr)
        if collections.Counter(direction)[1] > 1:  # diagonal
          line_id = tuple(x - rattr_idx[0] for x in rattr_idx[1:])
        else:
          line_id = rattr_idx[1:]
        reuse_groups[line_id][0].append((rattr_idx, aattr))
      _logger.debug('step 1')
      log_reuse_group()

      # step 2: sorting
      for line_id, (group, reuse_distance_list) in reuse_groups.items():
        group.sort(key=lambda attr: tuple(reversed(attr[0])), reverse=True)
        for rattr_idx, aattr in group:
          reuse_distance_list.append(group[0][0][0] - rattr_idx[0])
      _logger.debug('step 2')
      log_reuse_group()

      # step 3: find inner-group redundancy
      InnerGroupReuseKeyType = Tuple[int, Index, Index, Tuple[Any, ...]]
      InnerGroupReuseValueType = List[Tuple[List[Tuple[Index, Any]], Index,
                                            Optional[CommSchedule],
                                            List[Tuple[int, Any]]]]

      # dict mappping (stride, reused_distances, not_reused_distances, aattrs)
      # to list of (attrs, line_id, reused_schedule, new_attrs)
      inner_group_reuses = collections.defaultdict(
          list)  # type: Dict[InnerGroupReuseKeyType, InnerGroupReuseValueType]

      def log_inner_group_reuses(logger=_logger.debug,
                                 inner_group_reuses=inner_group_reuses) -> None:
        for (stride, reused, not_reused,
             aattrs), groups in inner_group_reuses.items():
          logger('logging inner group reuses')
          logger(
              '  stride: %d; reused distances: %s; not reused: %s; aattrs: %s',
              stride, reused, not_reused, util.lst2str(aattrs))
          for attrs, line_id, reused_schedule, new_attrs in groups:
            logger('    attrs: %s; line: %s; schedule: %s; new attrs: %s',
                   attrs, line_id, reused_schedule, new_attrs)
        logger('')

      for line_id, (group, reuse_distance_list) in reuse_groups.items():
        if len(group) > 3:
          reuse_dict = {}
          for stride in range(reuse_distance_list[1], reuse_distance_list[-1]):
            distance_map = {
                dist: attr for attr, dist in zip(group, reuse_distance_list)
            }
            distances = list(reuse_distance_list)
            reused_lst = []
            not_reused_lst = []
            new_attrs = []
            while distances:
              dist = distances.pop(0)
              if dist + stride in distances and (
                  distance_map[dist][1],
                  distance_map[dist + stride][1]) == (distance_map[0][1],
                                                      distance_map[stride][1]):
                distances.remove(dist + stride)
                reused_lst.append(dist)
                left_attr = (self.linearizer(distance_map[stride][0]),
                             distance_map[stride][1])
                right_attr = (self.linearizer(distance_map[0][0]),
                              distance_map[0][1])
                reused_schedule = linear_schedule((left_attr, right_attr))
                new_attrs.append(
                    (self.linearizer(distance_map[dist + stride][0]),
                     reused_schedule))
              else:
                not_reused_lst.append(dist)
                new_attrs.append((self.linearizer(distance_map[dist][0]),
                                  distance_map[dist][1]))
            reuse_dict[stride] = tuple(reused_lst), tuple(
                not_reused_lst), reused_schedule, new_attrs
          # best stride is the one with most reused items and least stride
          if reuse_dict:
            _logger.debug('reuse dict: %s', reuse_dict)
            stride, (reused, not_reused, reused_schedule,
                     new_attrs) = max(reuse_dict.items(),
                                      key=lambda item:
                                      (len(item[1][0]), -item[0]))
            new_attrs.sort(key=lambda attr: attr[0])
            _, aattrs = zip(*new_attrs)
            inner_group_reuses[(stride, reused, not_reused, aattrs)].append(
                (group, line_id, reused_schedule, new_attrs))
            continue
        _logger.debug('no reuse found')
        _, aattrs = zip(*reversed(group))
        inner_group_reuses[(0, (), tuple(reuse_distance_list), aattrs)].append(
            (group, line_id, None, [(self.linearizer(rattr_idx), aattr)
                                    for rattr_idx, aattr in reversed(group)]))
      _logger.debug('step 3')
      log_inner_group_reuses()

      # step 4: find inter-group redundancy
      all_attrs = []  # type: List[Tuple[int, Optional[CommSchedule]]]
      for (stride, reused, not_reused, _), groups in inner_group_reuses.items():
        if len(groups) > 1 and len(reused) + len(not_reused) > 1:  # with reuse
          groups.sort(key=lambda item: item[1])
          group, line_id, optional_reused_schedule, new_attrs = groups[0]
          _logger.debug('groups[0]: %s', groups[0])
          reused_expr = linear_schedule(new_attrs)
          for group, line_id, optional_reused_schedule, new_attrs in groups:
            _logger.debug('reusing %s @ %d', reused_expr, new_attrs[0][0])
            all_attrs.append((new_attrs[0][0], reused_expr))
          for group, line_id, optional_reused_schedule, new_attrs in groups:
            _logger.debug('reuse distance list: %s', reuse_groups[line_id][1])
            _logger.debug(
                'direction id: %s, stride: %s, reused: %s, not_reused: %s',
                line_id, stride, reused, not_reused)
            break
        else:  # without reuse
          for group, line_id, optional_reused_schedule, new_attrs in groups:
            all_attrs.extend(new_attrs)
      all_attrs.sort(key=lambda attr: attr[0])
      _logger.debug('attrs: %s', util.lst2str(map(util.idx2str, all_attrs)))
      schedule = linear_schedule(all_attrs)
      _logger.debug('schedule: %s (%d)', schedule, schedule.num_ops)
      yield schedule

  @cached_property.cached_property
  def best(self) -> CommSchedule:
    return min(self.generator, key=lambda schedule: schedule.num_ops)

  def print_stats(self, logger: Callable[..., None] = _logger.info) -> None:
    return


class ExternalSchedules(ScheduleBase):
  """Schedules of an Expression, found externally.
  """

  def __init__(self,
               rattrs: Tuple[RelativeAttr, ...],
               aattrs: Optional[Tuple[Union[AbsoluteAttr, CommSchedule, None],
                                      ...]] = None,
               linearizer: Optional[Linearizer] = None,
               cache: Optional[Dict] = None) -> None:
    self.linearizer = linearizer
    super().__init__(rattrs, aattrs)  # type: ignore
    self.cmd = ['soda-cr']

  def from_json(self, j: Dict[str, Any]) -> CommSchedule:
    return make_schedule_from_json(j, j['rattrs'][0], self.aattrs is None)

  @cached_property.cached_property
  def best(self) -> CommSchedule:
    attrs = {
        'rattrs': self.rattrs,
        'aattrs': self.aattrs or [1] * len(self.rattrs)
    }  # type: Dict[str, Any]
    if self.linearizer is not None and len(self.rattrs) >= 32:
      attrs['linearizer'] = {
          'maxs': self.linearizer.maxs,
          'mins': self.linearizer.mins,
          'sizes': self.linearizer.sizes,
      }
    if len(self.rattrs) < 32:
      attrs['num_pruned'] = 64
    elif len(self.rattrs) < 64:
      attrs['num_pruned'] = 4
    elif len(self.rattrs) < 128:
      attrs['num_pruned'] = 3
    elif len(self.rattrs) < 256:
      attrs['num_pruned'] = 2
    else:
      attrs['num_pruned'] = 1
    jemalloc = ctypes_util.find_library('jemalloc')
    if jemalloc is not None:
      os.environ['LD_PRELOAD'] = jemalloc
    result = json.loads(
        subprocess.run(self.cmd,
                       input=json.dumps(attrs),
                       stdout=subprocess.PIPE,
                       universal_newlines=True,
                       check=True).stdout)
    return self.from_json(result)

  def print_stats(self, logger: Callable[..., None] = _logger.info) -> None:
    pass


Schedule = CommSchedule
Schedules = Union[CommSchedules, GreedySchedules, GloreSchedules, BeamSchedules,
                  ExternalSchedules]


class Expression:
  """An expression suitable for computation reuse.

  Attributes:
    operator: String of the operator.
    operands: Tuple of all operands.
    rattrs: Tuple of relative attributes as integer offsets.
    aattrs: Tuple of absolute attributes as integer tags, or None.
    aattrs_as_ir_nodes: Tuple of absolute attributes as IR nodes.
    linearizer: A linearizer mapping relative attributes to tuples.
    aattr_table: A dict mapping absolute attributes to IR nodes.
  """

  @staticmethod
  def set_optimizations(optimizations: Sequence[str] = (
      'reorder-exploration', 'skip-with-partial-cost',
      'lazy-evaluation')) -> None:
    CommSchedules.set_optimizations(optimizations)

  class CannotHandle(Exception):

    def __init__(self, msg, details: str = '') -> None:
      details = details or (': ' + str(details))
      super().__init__('cannot handle ' + str(msg) + ' yet' + str(details))

  def __init__(self, polynomial: ir.Node, stencil) -> None:
    """Figure out whether a ir.Node is suitable for computation reuse.

    Construct a CR Expression of the input polynomial. If it cannot be handled
    but is a valid ir.Node instance, it raises Expression.CannotHandle so that
    the recursive visitor can continue to find polynomials.

    Args:
      polynomial: ir.BinaryOp to work with.
      method: One of 'yes', 'greedy', 'optimal'.

    Raises:
      Expression.CannotHandle: If the input cannot be handled but is not error.
      TypeError: If the input is not an instance of ir.Node.
    """
    self.method = stencil.optimizations.get('computation-reuse') or 'greedy'
    reduction = ir.to_reduction(polynomial)
    if reduction is not None:
      self.operator = reduction[0]
      for operand in reduction[1]:
        if len(soda.visitor.get_load_set(operand)) > 1:
          raise Expression.CannotHandle('multi-index operands', str(operand))
        if not soda.visitor.get_load_set(operand):
          raise Expression.CannotHandle('const operand', str(operand))
      self.operands = tuple(
          sorted(reduction[1],
                 key=lambda x: tuple(
                     reversed(soda.visitor.get_load_set(x)[0].idx))))
      rattrs, aattrs = zip(*map(extract_attr, self.operands))
      self.aattrs_as_ir_nodes = aattrs

      self.linearizer = Linearizer(rattrs, stencil.tile_size)

      # linearize the relative attribtues and memorize the mapping
      self.rattrs = tuple(map(self.linearizer, rattrs))

      # linearize the absolute attributes
      if len(set(aattrs)) == 1:
        self.aattrs = None  # type: Optional[Tuple[int, ...]]
        self.aattr_table = {None: aattrs[0]} \
            # type: Dict[Optional[int], ir.Node]

      else:
        tag = 0
        operand_table = {}  # type: Dict[ir.Node, int]
        self.aattr_table = {}
        for aattr in aattrs:
          if aattr not in operand_table:
            operand_table[aattr] = tag
            self.aattr_table[tag] = aattr
            tag += 1
        self.aattrs = tuple(operand_table[aattr] for aattr in aattrs)

      _logger.debug('polynomial: %s%s', self.operator,
                    util.idx2str(self.operands))
      _logger.debug('rattrs: %s', util.idx2str(self.rattrs))
      _logger.debug('aattrs: %s', util.idx2str(self.aattrs_as_ir_nodes))
    elif isinstance(polynomial, ir.Node):
      raise Expression.CannotHandle(type(polynomial).__name__)
    else:
      raise TypeError('expect an instance of ir.BinaryOp')

  @cached_property.cached_property
  def schedules(self) -> Schedules:
    external_cr = shutil.which('soda-cr')
    args = self.rattrs, self.aattrs, self.linearizer
    if self.method == 'glore':
      return GloreSchedules(*args)
    if external_cr is None or self.method.startswith('built-in'):
      if self.method.endswith('optimal'):
        return CommSchedules(self.rattrs, self.aattrs, cache={})
      if self.method.endswith('greedy'):
        return GreedySchedules(*args)
      return BeamSchedules(*args)
    external_schedules = ExternalSchedules(*args)
    if self.method == 'optimal':
      external_schedules.cmd.append('--brute-force')
    elif self.method == 'greedy':
      external_schedules.cmd.append('--greedy')
    elif self.method == 'beam':
      external_schedules.cmd.append('--beam')
    return external_schedules

  @cached_property.cached_property
  def best_schedule(self) -> Schedule:
    return self.schedules.best.bind_expression(self)



import logging

from haoda import util

logger = logging.getLogger().getChild(__name__)

def print_code(stencil, header_file):
  logger.info('generate host header code as %s' % header_file.name)
  printer = util.CppPrinter(header_file)
  println = printer.println
  do_indent = printer.do_indent
  un_indent = printer.un_indent
  println('#ifndef HALIDE_%s_H_' % stencil.app_name.upper())
  println('#define HALIDE_%s_H_' % stencil.app_name.upper())
  println()

  println('#ifndef HALIDE_ATTRIBUTE_ALIGN')
  do_indent()
  println('#ifdef _MSC_VER')
  do_indent()
  println('#define HALIDE_ATTRIBUTE_ALIGN(x) __declspec(align(x))')
  un_indent()
  println('#else')
  do_indent()
  println('#define HALIDE_ATTRIBUTE_ALIGN(x) __attribute__((aligned(x)))')
  un_indent()
  println('#endif')
  un_indent()
  println('#endif//HALIDE_ATTRIBUTE_ALIGN')
  println()

  println('#ifndef BUFFER_T_DEFINED')
  println('#define BUFFER_T_DEFINED')
  println('#include<stdbool.h>')
  println('#include<stdint.h>')
  println('typedef struct buffer_t {')
  do_indent()
  println('uint64_t dev;')
  println('uint8_t* host;')
  println('int32_t extent[4];')
  println('int32_t stride[4];')
  println('int32_t min[4];')
  println('int32_t elem_size;')
  println('HALIDE_ATTRIBUTE_ALIGN(1) bool host_dirty;')
  println('HALIDE_ATTRIBUTE_ALIGN(1) bool dev_dirty;')
  println('HALIDE_ATTRIBUTE_ALIGN(1) uint8_t _padding[10 - sizeof(void *)];')
  un_indent()
  println('} buffer_t;')
  println('#endif//BUFFER_T_DEFINED')
  println()

  println('#ifndef HALIDE_FUNCTION_ATTRS')
  println('#define HALIDE_FUNCTION_ATTRS')
  println('#endif//HALIDE_FUNCTION_ATTRS')
  println()

  tensors = stencil.input_names + stencil.output_names + stencil.param_names
  println('int {}({}const char* xclbin) HALIDE_FUNCTION_ATTRS;'.format(
      stencil.app_name,
      ''.join(map('buffer_t *var_{}_buffer, '.format, tensors))))
  println()

  println('#endif//HALIDE_%s_H_' % stencil.app_name.upper())
  println()



import json
import logging
import math
import os
import subprocess

from soda import core
from soda import dataflow
from soda.model import resource

logger = logging.getLogger().getChild(__name__)

record = {}

def accumulate_fifo(estimation_routed, width, depth):
  fifo_size = depth*width
  if fifo_size > 1024:
    inc = 16*1024//width
    modulo_depth = math.ceil(depth/inc)*inc
    query_proc = subprocess.Popen([
      os.environ['XILINX_BRAM_MODEL_PATH'] + '/query',
      str(width), str(modulo_depth)], stdout=subprocess.PIPE, bufsize=0)
    delta = json.loads(query_proc.stdout.read())
    delta = (delta['fifo_w%d_d%d' % (width, modulo_depth)]['BRAM_18K'] +
         delta['fifo_w%d_d%d' % (width, modulo_depth)]['BRAM_36K']*2)

    estimation_routed['BRAM'] += delta
    record.setdefault((width, depth, delta), 0)
    record[(width, depth, delta)] += 1
  else:
    estimation_routed['REG'] += width
    estimation_routed['LUT'] += depth/2

def print_estimation(stencil, model_file, output_file):
  model = json.load(model_file)
  if stencil.app_name not in model['compute']:
    msg = ('stencil app %s not found in model %s' %
      (stencil.app_name, model_file.name))
    raise core.SemanticError(msg)
  if model['vendor'] != 'xilinx':
    msg = 'cannot find a model for Xilinx in model %s' % model_file.name
    raise core.SemanticError(msg)
  estimation = {
    'resource_hls': {
      'BRAM_18K': 0,
      'DSP48E': 0,
      'FF': 0,
      'LUT': 0
    },
    'resource_routed': {
      'BRAM': 0,
      'DSP': 0,
      'LUT': 0,
      'LUTMem': 0,
      'REG': 0
    },
  }
  estimation_hls = estimation['resource_hls']

  # these modules appear and only appear once
  for module in ('Block_proc', 'control_s_axi', 'entry', 'top_level'):
    resource.accum_res(estimation_hls, model[module])

  # axi master
  resource.accum_res(estimation_hls, model['m_axi'],
             stencil.dram_bank*2)

  # load/store
  resource.accum_res(estimation_hls, model['load'],
             stencil.input.chan*stencil.dram_bank)
  resource.accum_res(estimation_hls, model['store'],
             stencil.output.chan*stencil.dram_bank)

  # unpack/pack
  resource.accum_res(estimation_hls,
             model['pack_'+stencil.output.type]['base'],
             stencil.output.chan*stencil.dram_bank)
  resource.accum_res(estimation_hls,
             model['pack_'+stencil.output.type]['directly'],
             stencil.output.chan*stencil.unroll_factor)
  resource.accum_res(estimation_hls,
             model['pack_'+stencil.output.type]['inversely'],
             stencil.output.chan/stencil.unroll_factor)
  resource.accum_res(estimation_hls,
             model['unpack_'+stencil.input.type]['base'],
             stencil.input.chan*stencil.dram_bank)
  resource.accum_res(estimation_hls,
             model['unpack_'+stencil.input.type]['directly'],
             stencil.input.chan*stencil.unroll_factor)
  resource.accum_res(estimation_hls,
             model['unpack_'+stencil.input.type]['inversely'],
             stencil.input.chan/stencil.unroll_factor)

  # modules in the dataflow graph
  super_source = stencil.dataflow_super_source
  if stencil.replication_factor is not None:
    repli = stencil.replication_factor
  else:
    repli = 1
  for node in super_source.bfs_node_generator():
    # forward modules
    if isinstance(node, dataflow.ForwardNode):
      forward_key = 'forward_'+node.tensor.type
      if forward_key not in model:
        msg = 'forward<%s> not found in model %s' % (node.tensor.type,
                               model_file.name)
        raise core.SemanticError(msg)
      resource.accum_res(estimation_hls, model[forward_key]['base'],
                 repli)
      resource.accum_res(estimation_hls,
                 model[forward_key]['overhead_per_dst'],
                 len(node.children)*repli)
      if node.depth > 0:
        resource.accum_res(estimation_hls,
                   model[forward_key]['overhead_of_depth'],
                   repli)
    elif isinstance(node, dataflow.ComputeNode):
      if node.stage.name in (stencil.output.name,):
        stage_name = node.stage.name
      elif stencil.input.name+'_iter' in node.stage.name:
        stage_name = stencil.output.name
      elif node.stage.name.split('_iter')[0] in stencil.tensors:
        stage_name = node.stage.name.split('_iter')[0]
      else:
        msg = 'unknown dataflow compute node: %s' % node.stage.name
        raise core.SemanticError(msg)
      resource.accum_res(estimation_hls,
                 model['compute'][stencil.app_name][stage_name],
                 repli)

  performance = stencil.unroll_factor*model['target_freq']/10**3  # pixel/ns
  dram_bandwidth = model['dram_bandwidth']
  if stencil.dram_separate:
    performance = min(performance,
      dram_bandwidth*stencil.dram_bank/
        (stencil.input.chan*core.TYPE_WIDTH[stencil.input.type]/8),
      dram_bandwidth*stencil.dram_bank/
        (stencil.output.chan*core.TYPE_WIDTH[stencil.output.type]/8))
  else:
    performance = min(performance,
      dram_bandwidth*stencil.dram_bank/(
        stencil.input.chan*core.TYPE_WIDTH[stencil.input.type]/8+
        stencil.output.chan*core.TYPE_WIDTH[stencil.output.type]/8))
  performance *= stencil.iterate

  estimation_routed = estimation['resource_routed']

  estimation_routed['DSP'] = estimation_hls['DSP48E']
  estimation_routed['BRAM'] = estimation_hls['BRAM_18K']
  estimation_routed['REG'] = (estimation_hls['FF'] *
                model['ff_coeff'] +
                model['ff_base'])
  estimation_routed['LUT'] = (estimation_hls['LUT'] *
                model['lut_coeff'] +
                model['lut_base'])
  estimation_routed['LUTMem'] = estimation_hls['LUT']*model['lutmem_coeff']

  # BRAMs
  for src_node, dst_node in super_source.bfs_edge_generator():
    if isinstance(dst_node, dataflow.ForwardNode):
      for c in range(dst_node.tensor.chan*repli):
        accumulate_fifo(estimation_routed,
                core.TYPE_WIDTH[dst_node.tensor.type],
                max(dst_node.depth, 1))
    elif (isinstance(src_node, dataflow.ForwardNode) and
        isinstance(dst_node, dataflow.ComputeNode)):
      extra_depth = super_source.get_extra_depth((src_node, dst_node))
      for c in range(src_node.tensor.chan*repli):
        accumulate_fifo(estimation_routed,
                core.TYPE_WIDTH[src_node.tensor.type],
                max(extra_depth+1, 2))

  estimation_routed['BRAM'] = math.ceil(estimation_routed['BRAM']/2)

  available_hls = model['available_hls']
  available_routed = model['available_routed']
  permissible = (
    all(estimation_hls[resource] < available_hls[resource]
      for resource in estimation_hls) and
    all(estimation_routed[resource] < available_routed[resource]
      for resource in estimation_routed))
  confidence = lambda a, b: math.atan2(abs(a-b)*64, b)/math.pi*2  # .9 -> .9
  permissible_confidence = min(
    *[
      confidence(estimation_hls[resource], available_hls[resource])
      for resource in estimation_hls
    ], *[
      confidence(estimation_routed[resource], available_routed[resource])
      for resource in estimation_routed
    ])
  estimation['permissible'] = permissible
  estimation['permissible_confidence'] = permissible_confidence
  estimation['performance'] = performance
  estimation['performance_unit'] = 'pixel/ns'

  json.dump(estimation, output_file, indent=2, sort_keys=True)



import collections
import functools
import logging
import operator

from haoda import ir
from haoda import util
from haoda.ir import visitor

_logger = logging.getLogger().getChild(__name__)

def _print_interface(printer, kernel_name, inputs, outputs, super_source):
  """Prints the top-level module for the given arguments.

  Prints the top-level interfaces and sub-module instances with proper interface
  pragmas, hls::stream declarations and references, and module function calls.
  Currently only streaming applications are supported.

  Args:
    printer: CppPrinter to which the code is emitted.
    kernel_name: str, name of the kernel.
    inputs: Sequence of (name, c_type, bank, depth) tuples, specifies the m_axi
      input interfaces.
    outputs: Sequence of (name, c_type, bank, depth) tuples, specifies the m_axi
      output interfaces.
    super_source: SuperSourceNode of a DAG of HAODA nodes.
  """
  println = printer.println
  do_indent = printer.do_indent
  un_indent = printer.un_indent
  do_scope = printer.do_scope
  un_scope = printer.un_scope

  get_bundle_name = util.get_bundle_name
  get_port_name = util.get_port_name
  get_port_buf_name = util.get_port_buf_name

  println('extern "C"')
  println('{')
  println()
  println('void %s(' % kernel_name)
  do_indent()
  for name, c_type, bank, _ in outputs + inputs:
    println('{}* {},'.format(c_type, get_port_name(name, bank)))
  println('uint64_t coalesced_data_num)')
  un_indent()
  do_scope()

  for name, c_type, bank, depth in outputs + inputs:
    println('#pragma HLS interface m_axi port={} offset=slave depth={} bundle={'
            '}'.format(get_port_name(name, bank), depth,
                       get_bundle_name(name, bank)), 0)

  println()
  for name, _, bank, _ in outputs + inputs:
    println('#pragma HLS interface s_axilite port={} bundle=control'.format(
        get_port_name(name, bank)), 0)

  println('#pragma HLS interface s_axilite port=coalesced_data_num '
          'bundle=control', 0)
  println('#pragma HLS interface s_axilite port=return bundle=control', 0)
  println()

  # port buf declarations
  for name, c_type, bank, _ in inputs + outputs:
    println('hls::stream<Data<{c_type}>> {name}("{name}");'.format(
        name=get_port_buf_name(name, bank), c_type=c_type))
  # port buf depths
    println('#pragma HLS stream variable={} depth=32'.format(
        get_port_buf_name(name, bank)), 0)
    println('#pragma HLS data_pack variable={}'.format(
        get_port_buf_name(name, bank)), indent=0)
  println()

  # internal fifos
  for node in super_source.tpo_valid_node_gen():
    for fifo in node.fifos:
      println('hls::stream<Data<{0}>> {1}("{1}");'.format(fifo.c_type,
                                                          fifo.c_expr))
      println('#pragma HLS stream variable={} depth={}'.format(
          fifo.c_expr, fifo.depth + 3), 0)
      println('#pragma HLS data_pack variable={}'.format(fifo.c_expr),
              indent=0)

  # pylint: disable=pointless-string-statement
  '''
  if extra_params:
    for param in extra_params.values():
      if param.dup:
        dup = ('[%d]' % param.dup)
      else:
        dup = ''
      println('%s %s%s[UNROLL_FACTOR][%s];' % (param.type, param.name, dup, ']['.join(map(str, param.size))))
    println()

    for param in extra_params.values():
      println('#pragma HLS array_partition variable=%s complete dim=1' % param.name, 0)
      dim_offset = 1
      if param.dup:
        println('#pragma HLS array_partition variable=%s complete dim=2' % param.name, 0)
        dim_offset = 2
      for partitioning in param.partitioning:
        println('#pragma HLS array_partition variable=%s %s dim=%d%s' % (
          param.name,
          partitioning.partition_type,
          dim_offset+1 if partitioning.dim is None else partitioning.dim+dim_offset,
          '' if partitioning.factor is None else ' factor=%d' % partitioning.factor,
        ), 0)
    println()

    for param in extra_params.values():
      if len(param.size) > 1:
        for dim, size in enumerate(param.size):
          println('uint32_t %s_index_dim_%d = 0;' % (param.name, dim))
      println('%s_init:' % param.name, 0)
      println('for(int %s_index = 0; %s_index < %d; ++%s_index)' % (param.name, param.name, functools.reduce(operator.mul, param.size), param.name))
      p.do_scope()
      println('#pragma HLS pipeline II=1', 0)
      println('%s& %s_tmp = var_%s[%s_index];' % (param.type, param.name, param.name, param.name))
      println('%s_unrolled:' % param.name, 0)
      println('for(int unroll_index = 0; unroll_index < UNROLL_FACTOR; ++unroll_index)')
      p.do_scope()
      println('#pragma HLS unroll',0)
      if param.dup is None:
        println('%s[unroll_index]%s = %s_tmp;' % ((param.name, ''.join(['[%s_index_dim_%d]' % (param.name, x) for x in range(len(param.size)-1, -1, -1)]) if len(param.size)>1 else '[%s_index]' % param.name, param.name)))
      else:
        for i in range(param.dup):
          println('%s[%d][unroll_index]%s = %s_tmp;' % (param.name, i, ''.join(['[%s_index_dim_%d]' % (param.name, x) for x in range(len(param.size)-1, -1, -1)]) if len(param.size)>1 else '[%s_index]' % param.name, param.name))
      p.un_scope()
      if len(param.size) > 1:
        for dim in range(len(param.size)):
          println('++%s_index_dim_%d;' % (param.name, dim))
          if dim<len(param.size)-1:
            println('if(%s_index_dim_%d==%d)' % (param.name, dim, param.size[len(param.size)-1-dim]))
            p.do_scope()
            println('%s_index_dim_%d = 0;' % (param.name, dim))
      for size in param.size[:-1]:
        p.un_scope()
      p.un_scope()
    println()

  extra_params_str = ''.join([param.name+', ' for param in extra_params.values()])
  '''

  println()

  # TODO: replication not supported
  # pylint: disable=pointless-string-statement
  '''
  if stencil.replication_factor > 1:
    _generate_code(p, stencil)

    p.un_scope()
    println()
    println('}  //extern "C"')

    return

  # reuse buffers
  if stencil.cluster == 'none':
    for name, reuse_buffer in reuse_buffers.items():
      pragmas = []
      msg = 'reuse buffers for %s' % name
      _logger.debug('generate %s', msg)
      println('// %s' % msg)
      for start, end in reuse_buffer[1:]:
        println('hls::stream<{0}> {1}("{1}");'.format(
            stencil.tensors[name].c_type, get_tensor_name_at(name, end)))
        buffer_length = stencil.reuse_buffer_lengths[name][end]
        tensor_name = get_tensor_name_at(name, end)
        if buffer_length > 1:
          pragmas.append((tensor_name, buffer_length))
      for pragma in pragmas:
        println('#pragma HLS stream variable=%s depth=%d' % pragma, 0)
      println()
  else:
    println('// %s' % stencil.input.name)
    for unroll_index in range(unroll_factor):
      for c in range(stencil.input.chan):
        println('hls::stream<{0}> {1}("{1}");'.format(
            stencil.tensors[stencil.input.name].type,
            get_tensor_name_at(stencil.input.name, unroll_index, c)))
    println()

  for output_name in stencil.output_names:
    println('// %s' % output_name)
    for unroll_index in range(unroll_factor):
      println('hls::stream<{0}> {1}("{1}");'.format(
          stencil.tensors[output_name].c_type,
          get_tensor_name_at(output_name, unroll_index)))
  println()

  # params
  msg = 'params'
  _logger.debug('generate %s' % msg)
  println('// %s' % msg)
  pragmas = []
  for tensor in stencil.chronological_tensors:
    for pe_id in range(unroll_factor):
      for input_name, input_window in tensor.ld_indices.items():
        for param_id in range(len(input_window)):
          offset = next(offset for offset, points in
            all_points[input_name][tensor.name].items()
            if pe_id in points and points[pe_id] == param_id)
          fwd_node = super_source.fwd_nodes[(input_name, offset)]
          cpt_node = super_source.cpt_nodes[(tensor.name, pe_id)]
          extra_depth = super_source.get_extra_depth(
            (fwd_node, cpt_node))
          var_type = stencil.tensors[input_name].c_type
          var_name = 'from_%s_to_%s_param_%d_pe_%d' % (
            input_name, tensor.name, param_id, pe_id)
          println('hls::stream<{0}> {1}("{1}");'.format(var_type, var_name))
          if extra_depth > 0:
            pragmas.append((var_name, extra_depth+1))
          else:
            pragmas.append((var_name, 2))
  for pragma in pragmas:
    println('#pragma HLS stream variable=%s depth=%d' % pragma, 0)
  println()
  '''

  '''
  # border buffers
  msg = 'border buffers'
  _logger.debug('generate %s' % msg)
  println('// %s' % msg)
  for tensor in stencil.chronological_tensors:
    if tensor.output.preserve_border_to():
      for unroll_index in range(unroll_factor):
        for d in range(stencil.dim-1):
          for c in range(tensor.output.chan):
            println('hls::stream<{0}> {1}("{1}");'.foramt(
                tensor.output.type,
                'border_from_%s_dim_%d_left_chan_%d_pe_%d' % (
                    tensor.name, d, c, unroll_index)))
            println('hls::stream<{0}> {1}("{1}");'.foramt(
                tensor.output.type,
                'border_from_%s_dim_%d_right_chan_%d_pe_%d' % (
                    tensor.name, d, c, unroll_index)))
  println()
  '''

  println('#pragma HLS dataflow', 0)
  for name, _, bank, _ in inputs:
    println('BurstRead(&{}, {}, coalesced_data_num);'.format(
        get_port_buf_name(name, bank), get_port_name(name, bank)))

  for node in super_source.tpo_valid_node_gen():
    module_trait_id = super_source.module_table[node][1]
    _print_module_func_call(printer, node, module_trait_id)

  for name, _, bank, _ in outputs:
    println('BurstWrite({}, &{}, coalesced_data_num);'.format(
        get_port_name(name, bank), get_port_buf_name(name, bank)))

  un_scope()
  println()
  println('}//extern "C"')

def print_header(printer):
  println = printer.println
  for header in ['float', 'math', 'stdbool', 'stddef', 'stdint', 'stdio',
                 'string', 'ap_int', 'hls_stream']:
    println('#include<%s.h>' % header)
  println()
  for header in ['algorithm']:
    println('#include<%s>' % header)
  println()

def _print_burst_read(printer):
  println = printer.println
  do_scope = printer.do_scope
  un_scope = printer.un_scope
  println('void BurstRead(hls::stream<Data<ap_uint<BURST_WIDTH>>>* to, ap_uint<'
          'BURST_WIDTH>* from, uint64_t data_num)')
  do_scope()
  println('load_epoch:', 0)
  println('for (uint64_t epoch = 0; epoch < data_num;)')
  do_scope()
  println('#pragma HLS pipeline II=1', 0)
  println('const uint64_t next_epoch = epoch + 1;')
  println('WriteData(to, from[epoch], next_epoch < data_num);')
  println('epoch = next_epoch;')
  un_scope()
  un_scope()

def _print_burst_write(printer):
  println = printer.println
  do_scope = printer.do_scope
  un_scope = printer.un_scope
  println('void BurstWrite(ap_uint<BURST_WIDTH>* to, hls::stream<Data<ap_uint<B'
          'URST_WIDTH>>>* from, uint64_t data_num)')
  do_scope()
  println('store_epoch:', 0)
  println('for (uint64_t epoch = 0; epoch < data_num; ++epoch)')
  do_scope()
  println('#pragma HLS pipeline II=1', 0)
  println('ap_uint<BURST_WIDTH> buf;')
  println('ReadData(&buf, from);')
  println('to[epoch] = buf;')
  un_scope()
  un_scope()

def print_code(stencil, output_file):
  _logger.info('generate kernel code as %s' % output_file.name)
  printer = util.CppPrinter(output_file)

  print_header(printer)

  printer.println()

  util.print_define(printer, 'BURST_WIDTH', stencil.burst_width)
  printer.println()

  util.print_guard(printer, 'UNROLL_FACTOR', stencil.unroll_factor)
  for i in range(len(stencil.tile_size)-1):
    util.print_guard(printer, 'TILE_SIZE_DIM_%d' % i, stencil.tile_size[i])
  util.print_guard(printer, 'BURST_WIDTH', stencil.burst_width)
  printer.println()

  _print_data_struct(printer)
  _print_reinterpret(printer)
  _print_read_data(printer)
  _print_write_data(printer)

  _print_burst_read(printer)
  _print_burst_write(printer)

  for module_trait_id, module_trait in enumerate(stencil.module_traits):
    print_module_definition(printer, module_trait, module_trait_id,
                           burst_width=stencil.burst_width)


  # pylint: disable=pointless-string-statement
  '''
  if stencil.cluster == 'none':
    for forwarder in stencil.get_forwarders():
      print_forward_func(printer, forwarder)
      printer.println()

    for forwarder in stencil.get_forwarders_with_border():
      print_forward_func_with_border(printer, stencil, forwarder)
      printer.println()
    for stage in stencil.stages.values():
      print_compute_stage(printer, stencil, stage)
      printer.println()
  elif stencil.cluster == 'fine':
    print_compute_input(printer, stencil)
    for stage in stencil.get_stages_chronologically():
      print_compute_stage(printer, stencil, stage)

  if stencil.replication_factor > 1:
    dst_lists = set()
    super_source = stencil.dataflow_super_source
    def add_dst_lists(tensor):
      rf = stencil.replication_factor
      dst_ids = [start%rf for start, end
        in stencil.get_replicated_reuse_buffers()[tensor.name][1:]
        if start == end]
      dst_ids = tuple(dst_id if dst_id in dst_ids else None
        for dst_id in range(stencil.replication_factor))
      dst_lists.add(dst_ids)
    for node in super_source.tpo_node_generator():
      if isinstance(node, SuperSourceNode):
        add_dst_lists(stencil.input)
      elif isinstance(node, ComputeNode):
        if not node.stage.is_output():
          add_dst_lists(node.stage.output)
    for dst_list in dst_lists:
      _print_reconnect_func(printer, dst_list)
  printer.println()
  '''

  outputs = []
  inputs = []
  for stmt in stencil.output_stmts:
    for bank in stmt.dram:
      outputs.append((stmt.name, 'ap_uint<%d>' % stencil.burst_width, bank,
                      65536))
  for stmt in stencil.input_stmts:
    for bank in stmt.dram:
      inputs.append((stmt.name, 'ap_uint<%d>' % stencil.burst_width, bank,
                     65536))
  for stmt in stencil.param_stmts:
    inputs.append(('var_%s' % stmt.name, stmt.type, 0,
                    functools.reduce(operator.mul, stmt.size)))
  _print_interface(printer, stencil.app_name + '_kernel', inputs, outputs,
                   stencil.dataflow_super_source)

def _print_module_func_call(printer, node, module_trait_id, **kwargs):
  println = printer.println
  print_func = printer.print_func
  func_name = util.get_func_name(module_trait_id)

  dram_reads = tuple(
      '/* input*/ &' + util.get_port_buf_name(dram_ref.var, bank)
      for dram_ref, bank in node.dram_reads)
  dram_writes = tuple(
      '/*output*/ &' + util.get_port_buf_name(dram_ref.var, bank)
      for dram_ref, bank in node.dram_writes)
  output_fifos = tuple('/*output*/ &' + _ for _ in node.output_fifos)
  input_fifos = tuple('/* input*/ &' + _ for _ in node.input_fifos)
  params = dram_writes + output_fifos + input_fifos + dram_reads

  print_func(func_name, params, suffix=';', align=0)

# pylint: disable=too-many-branches,too-many-statements
def print_module_definition(printer, module_trait, module_trait_id, **kwargs):
  println = printer.println
  do_scope = printer.do_scope
  un_scope = printer.un_scope
  func_name = util.get_func_name(module_trait_id)
  func_lower_name = util.get_module_name(module_trait_id)
  ii = 1

  def get_delays(obj, delays):
    if isinstance(obj, ir.DelayedRef):
      delays.append(obj)
    return obj
  delays = []
  for let in module_trait.lets:
    let.visit(get_delays, delays)
  for expr in module_trait.exprs:
    expr.visit(get_delays, delays)
  _logger.debug('delays: %s', delays)

  fifo_loads = tuple('/* input*/ hls::stream<Data<{}>>* {}'.format(
      _.c_type, _.ld_name) for _ in module_trait.loads)
  fifo_stores = tuple('/*output*/ hls::stream<Data<{}>>* {}{}'.format(
      expr.c_type, ir.FIFORef.ST_PREFIX, idx)
    for idx, expr in enumerate(module_trait.exprs))

  # look for DRAM access
  reads_in_lets = tuple(_.expr for _ in module_trait.lets)
  writes_in_lets = tuple(_.name for _ in module_trait.lets
                         if not isinstance(_.name, str))
  reads_in_exprs = module_trait.exprs
  dram_reads = visitor.get_dram_refs(reads_in_lets + reads_in_exprs)
  dram_writes = visitor.get_dram_refs(writes_in_lets)
  dram_read_map = collections.OrderedDict()
  dram_write_map = collections.OrderedDict()
  all_dram_reads = ()
  num_bank_map = {}
  if dram_reads:  # this is an unpacking module
    assert not dram_writes, 'cannot read and write DRAM in the same module'
    for dram_read in dram_reads:
      dram_read_map.setdefault(dram_read.var,
                               collections.OrderedDict()).setdefault(
                                   dram_read.dram, []).append(dram_read)
    _logger.debug('dram read map: %s', dram_read_map)
    burst_width = kwargs.pop('burst_width')
    for var in dram_read_map:
      for dram in dram_read_map[var]:
        # number of elements per cycle
        batch_size = len(dram_read_map[var][dram])
        dram_read_map[var][dram] = collections.OrderedDict(
            (_.offset, _) for _ in dram_read_map[var][dram])
        dram_reads = dram_read_map[var][dram]
        num_banks = len(next(iter(dram_reads.values())).dram)
        if var in num_bank_map:
          assert num_bank_map[var] == num_banks, 'inconsistent num banks'
        else:
          num_bank_map[var] = num_banks
        _logger.debug('dram reads: %s', dram_reads)
        assert tuple(sorted(dram_reads.keys())) == tuple(range(batch_size)), \
               'unexpected DRAM accesses pattern %s' % dram_reads
        batch_width = sum(_.haoda_type.width_in_bits
                          for _ in dram_reads.values())
        del dram_reads
        if burst_width * num_banks >= batch_width:
          assert burst_width * num_banks % batch_width == 0, \
              'cannot process such a burst'
          # a single burst consumed in multiple cycles
          coalescing_factor = burst_width * num_banks // batch_width
          ii = coalescing_factor
        else:
          assert batch_width * num_banks % burst_width == 0, \
              'cannot process such a burst'
          # multiple bursts consumed in a single cycle
          # reassemble_factor = batch_width // (burst_width * num_banks)
          raise util.InternalError('cannot process such a burst yet')
      dram_reads = tuple(next(iter(_.values()))
                         for _ in dram_read_map[var].values())
      all_dram_reads += dram_reads
      fifo_loads += tuple(
          '/* input*/ hls::stream<Data<ap_uint<{burst_width}>>>* '
          '{bank_name}'.format(
              burst_width=burst_width, bank_name=_.dram_fifo_name(bank))
          for _ in dram_reads for bank in _.dram)
  elif dram_writes:   # this is a packing module
    for dram_write in dram_writes:
      dram_write_map.setdefault(dram_write.var,
                                collections.OrderedDict()).setdefault(
                                    dram_write.dram, []).append(dram_write)
    _logger.debug('dram write map: %s', dram_write_map)
    burst_width = kwargs.pop('burst_width')
    for var in dram_write_map:
      for dram in dram_write_map[var]:
        # number of elements per cycle
        batch_size = len(dram_write_map[var][dram])
        dram_write_map[var][dram] = collections.OrderedDict(
            (_.offset, _) for _ in dram_write_map[var][dram])
        dram_writes = dram_write_map[var][dram]
        num_banks = len(next(iter(dram_writes.values())).dram)
        if var in num_bank_map:
          assert num_bank_map[var] == num_banks, 'inconsistent num banks'
        else:
          num_bank_map[var] = num_banks
        _logger.debug('dram writes: %s', dram_writes)
        assert tuple(sorted(dram_writes.keys())) == tuple(range(batch_size)), \
               'unexpected DRAM accesses pattern %s' % dram_writes
        batch_width = sum(_.haoda_type.width_in_bits
                          for _ in dram_writes.values())
        del dram_writes
        if burst_width * num_banks >= batch_width:
          assert burst_width * num_banks % batch_width == 0, \
              'cannot process such a burst'
          # a single burst consumed in multiple cycles
          coalescing_factor = burst_width * num_banks // batch_width
          ii = coalescing_factor
        else:
          assert batch_width * num_banks % burst_width == 0, \
              'cannot process such a burst'
          # multiple bursts consumed in a single cycle
          # reassemble_factor = batch_width // (burst_width * num_banks)
          raise util.InternalError('cannot process such a burst yet')
      dram_writes = tuple(next(iter(_.values()))
                          for _ in dram_write_map[var].values())
      fifo_stores += tuple(
          '/*output*/ hls::stream<Data<ap_uint<{burst_width}>>>* '
          '{bank_name}'.format(
              burst_width=burst_width, bank_name=_.dram_fifo_name(bank))
          for _ in dram_writes for bank in _.dram)

  # print function
  printer.print_func('void {func_name}'.format(**locals()),
                     fifo_stores+fifo_loads, align=0)
  do_scope(func_name)

  for dram_ref, bank in module_trait.dram_writes:
    println('#pragma HLS data_pack variable = {}'.format(
        dram_ref.dram_fifo_name(bank)), 0)
  for arg in module_trait.output_fifos:
    println('#pragma HLS data_pack variable = %s' % arg, 0)
  for arg in module_trait.input_fifos:
    println('#pragma HLS data_pack variable = %s' % arg, 0)
  for dram_ref, bank in module_trait.dram_reads:
    println('#pragma HLS data_pack variable = {}'.format(
        dram_ref.dram_fifo_name(bank)), 0)

  # print inter-iteration declarations
  for delay in delays:
    println(delay.c_buf_decl)
    println(delay.c_ptr_decl)

  # print loop
  println('{}_epoch:'.format(func_lower_name), indent=0)
  println('for (bool enable = true; enable;)')
  do_scope('for {}_epoch'.format(func_lower_name))
  println('#pragma HLS pipeline II=%d' % ii, 0)
  for delay in delays:
    println('#pragma HLS dependence variable=%s inter false' %
            delay.buf_name, 0)

  # print emptyness tests
  println('if (%s)' % (' && '.join(
      '!{fifo}->empty()'.format(fifo=fifo)
      for fifo in tuple(_.ld_name for _ in module_trait.loads) +
                  tuple(_.dram_fifo_name(bank)
                        for _ in all_dram_reads for bank in _.dram))))
  do_scope('if not empty')

  # print intra-iteration declarations
  for fifo_in in module_trait.loads:
    println('{fifo_in.c_type} {fifo_in.ref_name};'.format(**locals()))
  for var in dram_read_map:
    for dram in (next(iter(_.values())) for _ in dram_read_map[var].values()):
      for bank in dram.dram:
        println('ap_uint<{}> {};'.format(burst_width, dram.dram_buf_name(bank)))
  for var in dram_write_map:
    for dram in (next(iter(_.values())) for _ in dram_write_map[var].values()):
      for bank in dram.dram:
        println('ap_uint<{}> {};'.format(burst_width, dram.dram_buf_name(bank)))

  # print enable conditions
  if not dram_write_map:
    for fifo_in in module_trait.loads:
      println('const bool {fifo_in.ref_name}_enable = '
        'ReadData(&{fifo_in.ref_name}, {fifo_in.ld_name});'.format(**locals()))
  for dram in all_dram_reads:
    for bank in dram.dram:
      println('const bool {dram_buf_name}_enable = '
              'ReadData(&{dram_buf_name}, {dram_fifo_name});'.format(
                  dram_buf_name=dram.dram_buf_name(bank),
                  dram_fifo_name=dram.dram_fifo_name(bank)))
  if not dram_write_map:
    println('const bool enabled = %s;' % (
      ' && '.join(tuple('{_.ref_name}_enable'.format(_=_)
                        for _ in module_trait.loads) +
                  tuple('{}_enable'.format(_.dram_buf_name(bank))
                        for _ in all_dram_reads for bank in _.dram))))
    println('enable = enabled;')

  # print delays (if any)
  for delay in delays:
    println('const {} {};'.format(delay.c_type, delay.c_buf_load))

  # print lets
  def mutate_dram_ref_for_writes(obj, kwargs):
    if isinstance(obj, ir.DRAMRef):
      coalescing_idx = kwargs.pop('coalescing_idx')
      unroll_factor = kwargs.pop('unroll_factor')
      type_width = obj.haoda_type.width_in_bits
      elem_idx = coalescing_idx * unroll_factor + obj.offset
      num_banks = num_bank_map[obj.var]
      bank = obj.dram[elem_idx % num_banks]
      lsb = (elem_idx // num_banks) * type_width
      msb = lsb + type_width - 1
      return ir.Var(name='{}({msb}, {lsb})'.format(
          obj.dram_buf_name(bank), msb=msb, lsb=lsb), idx=())
    return obj

  # mutate dram ref for writes
  if dram_write_map:
    for coalescing_idx in range(coalescing_factor):
      for fifo_in in module_trait.loads:
        if coalescing_idx == coalescing_factor - 1:
          prefix = 'const bool {fifo_in.ref_name}_enable = '.format(
              fifo_in=fifo_in)
        else:
          prefix = ''
        println('{prefix}ReadData(&{fifo_in.ref_name},'
                ' {fifo_in.ld_name});'.format(fifo_in=fifo_in, prefix=prefix))
      if coalescing_idx == coalescing_factor - 1:
        println('const bool enabled = %s;' % (
          ' && '.join(tuple('{_.ref_name}_enable'.format(_=_)
                            for _ in module_trait.loads) +
                      tuple('{}_enable'.format(_.dram_buf_name(bank))
                            for _ in dram_reads for bank in _.dram))))
        println('enable = enabled;')
      for idx, let in enumerate(module_trait.lets):
        let = let.visit(mutate_dram_ref_for_writes, {
            'coalescing_idx': coalescing_idx, 'unroll_factor': len(
                dram_write_map[let.name.var][let.name.dram])})
        println('{} = Reinterpret<ap_uint<{width}>>({});'.format(
            let.name, let.expr.c_expr,
            width=let.expr.haoda_type.width_in_bits))
    for var in dram_write_map:
      for dram in (next(iter(_.values()))
                   for _ in dram_write_map[var].values()):
        for bank in dram.dram:
          println('WriteData({}, {}, enabled);'.format(
              dram.dram_fifo_name(bank), dram.dram_buf_name(bank)))
  else:
    for let in module_trait.lets:
      println(let.c_expr)

  def mutate_dram_ref_for_reads(obj, kwargs):
    if isinstance(obj, ir.DRAMRef):
      coalescing_idx = kwargs.pop('coalescing_idx')
      unroll_factor = kwargs.pop('unroll_factor')
      type_width = obj.haoda_type.width_in_bits
      elem_idx = coalescing_idx * unroll_factor + obj.offset
      num_banks = num_bank_map[obj.var]
      bank = expr.dram[elem_idx % num_banks]
      lsb = (elem_idx // num_banks) * type_width
      msb = lsb + type_width - 1
      return ir.Var(
          name='Reinterpret<{c_type}>(static_cast<ap_uint<{width}>>('
               '{dram_buf_name}({msb}, {lsb})))'.format(
                   c_type=obj.c_type, dram_buf_name=obj.dram_buf_name(bank),
                   msb=msb, lsb=lsb, width=msb-lsb+1), idx=())
    return obj

  # mutate dram ref for reads
  if dram_read_map:
    for coalescing_idx in range(coalescing_factor):
      for idx, expr in enumerate(module_trait.exprs):
        println('WriteData({}{}, {}, {});'.format(
            ir.FIFORef.ST_PREFIX, idx,
            expr.visit(mutate_dram_ref_for_reads, {
                'coalescing_idx': coalescing_idx, 'unroll_factor': len(
                    dram_read_map[expr.var][expr.dram])}).c_expr,
            'true' if coalescing_idx < coalescing_factor - 1 else 'enabled'))
  else:
    for idx, expr in enumerate(module_trait.exprs):
      println('WriteData({}{}, {}({}), enabled);'.format(
              ir.FIFORef.ST_PREFIX, idx, expr.c_type, expr.c_expr))

  for delay in delays:
    println(delay.c_buf_store)
    println('{} = {};'.format(delay.ptr, delay.c_next_ptr_expr))

  un_scope()
  un_scope()
  un_scope()
  _logger.debug('printing: %s', module_trait)

def _print_data_struct(printer):
  println = printer.println
  println('template<typename T> struct Data')
  printer.do_scope()
  println('T data;')
  println('bool ctrl;')
  printer.un_scope(suffix=';')

def _print_reinterpret(printer):
  println = printer.println
  println('template<typename To, typename From>')
  println('inline To Reinterpret(From val)')
  printer.do_scope()
  println('#pragma HLS inline', indent=0)
  println('return reinterpret_cast<To&>(val);')
  printer.un_scope()

def _print_read_data(printer):
  println = printer.println
  println('template<typename T> inline bool ReadData'
          '(T* data, hls::stream<Data<T>>* from)')
  printer.do_scope()
  println('#pragma HLS inline', indent=0)
  println('const Data<T>& tmp = from->read();')
  println('*data = tmp.data;')
  println('return tmp.ctrl;')
  printer.un_scope()

def _print_write_data(printer):
  println = printer.println
  println('template<typename T> inline void WriteData'
          '(hls::stream<Data<T>>* to, const T& data, bool ctrl)')
  printer.do_scope()
  println('#pragma HLS inline', indent=0)
  println('Data<T> tmp;')
  println('tmp.data = data;')
  println('tmp.ctrl = ctrl;')
  println('to->write(tmp);')
  printer.un_scope()



from typing import (
    Dict,
    Tuple,
)

import logging

import toposort

from haoda import ir, util
from haoda.ir.arithmetic import base

_logger = logging.getLogger().getChild(__name__)

GRAMMAR = r'''
SodaProgram:
(
  ('border' ':' border=BorderStrategies)?
  ('burst' 'width' ':' burst_width=INT)
  ('cluster' ':' cluster=ClusterStrategies)?
  ('iterate' ':' iterate=INT)
  ('kernel' ':' app_name=ID)
  ('unroll' 'factor' ':' unroll_factor=INT)
  (input_stmts=InputStmt)+
  (param_stmts=ParamStmt)*
  (local_stmts=LocalStmt)*
  (output_stmts=OutputStmt)+
)#;

YesOrNo: 'yes'|'no';

BorderStrategies: 'ignore'|'preserve';
ClusterStrategies: 'none'|'fine'|'coarse'|'full';

Comment: /\s*#.*$/;

InputStmt: 'input' ('dram' dram=INT ('.' dram=INT)*)? haoda_type=Type ':' name=ID ('(' (tile_size=INT ',')* '*' ')')?;
LocalStmt: 'local' haoda_type=Type ':' (let=Let)* ref=Ref '=' expr=Expr;
OutputStmt: 'output' ('dram' dram=INT ('.' dram=INT)*)? haoda_type=Type ':' (let=Let)* ref=Ref '=' expr=Expr;

ParamStmt: 'param' ('dram' dram=INT ('.' dram=INT)*)? haoda_type=Type (',' attr=ParamAttr)* ':' name=ID ('[' size=INT ']')*;
ParamAttr: 'dup' dup=Int | partitioning=Partitioning;
Partitioning:
  'partition' strategy='complete' ('dim' '=' dim=Int)? |
  'partition' strategy='cyclic' 'factor' '=' factor=Int ('dim' '=' dim=Int)?;
''' + ir.GRAMMAR

class InputStmt(ir.Node):
  """Node for input statement, represents a tiled input tensor.

  Attributes:
    haoda_type: Type of this input tensor.
    dram: [int], dram id used to read this input
    name: str, name of this input tensor.
    tile_size: list of tile sizes. The last dimension should be 0.
  """
  SCALAR_ATTRS = 'haoda_type', 'name'
  LINEAR_ATTRS = ('tile_size', 'dram',)
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    # pylint: disable=access-member-before-definition
    if not self.dram:
      self.dram = (0,)
    self.tile_size += (0,)

  def __str__(self):
    result = 'input {}: {}'.format(self.haoda_type, self.name)
    if self.tile_size[:-1]:
      result += '({}, *)'.format(', '.join(map(str, self.tile_size[:-1])))
    return result

class LocalStmtOrOutputStmt(ir.Node):
  SCALAR_ATTRS: Tuple[str, ...]
  LINEAR_ATTRS: Tuple[str, ...]
  SCALAR_ATTRS = 'haoda_type', 'ref', 'expr'
  LINEAR_ATTRS = ('let',)
  def __init__(self, **kwargs):
    # inform mypy of the attributes
    self.haoda_type = None
    self.ref = None
    self.expr = None
    self.let = ()
    super().__init__(**kwargs)
    var_types = {}
    # pylint: disable=access-member-before-definition
    for let in self.let:
      var_types[let.name] = let.haoda_type
    def set_var_type(obj, var_types):
      if isinstance(obj, ir.Var) and obj.name in var_types:
        obj.haoda_type = var_types[obj.name]
      return obj
    self.let = tuple(_.visit(set_var_type, var_types) for _ in self.let)
    self.expr = self.expr.visit(set_var_type, var_types)
    self.stencil = kwargs.pop('stencil', None)

  @property
  def name(self):
    return self.ref.name

  def __str__(self):
    if self.let:
      let = '\n  {}\n '.format('\n  '.join(map(str, self.let)))
    else:
      let = ''
    return '{} {}:{} {} = {}'.format(type(self).__name__[:-4].lower(),
                                     self.haoda_type, let, self.ref,
                                     ir.unparenthesize(self.expr))

  @property
  def symbol_table(self) -> Dict[str, str]:
    # types of lets are local to this statement
    # must **not** modify self.stencil.symbol_table in-place
    symbol_table = self.stencil.symbol_table.copy()
    lets = {let.name: let for let in self.let}
    for var in toposort.toposort_flatten({
        let.name: {var.name for var in ir.visitor.get_vars(let)}
        for let in self.let}):
      symbol_table[var] = base.propagate_type(lets[var],
                                              symbol_table).haoda_type
    return symbol_table


  def propagate_type(self, dummy=None) -> None:
    """Propagate haoda type of the nodes in this statement.

    Args:
      symbol_table: A dict mapping input or local tensor names to haoda types.

    Returns:
      None.
    """
    symbol_table = self.symbol_table
    self.expr = base.propagate_type(self.expr, symbol_table)
    if self.expr.haoda_type != self.haoda_type:
      self.expr = ir.Cast(expr=self.expr, haoda_type=self.haoda_type)
    self.let = tuple(base.propagate_type(let, symbol_table) for let in self.let)

class LocalStmt(LocalStmtOrOutputStmt):
  pass

class OutputStmt(LocalStmtOrOutputStmt):
  LINEAR_ATTRS = LocalStmtOrOutputStmt.LINEAR_ATTRS + ('dram',)
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    # pylint: disable=access-member-before-definition
    if not self.dram:
      self.dram = (0,)

class ParamStmt(ir.Node):
  SCALAR_ATTRS = 'haoda_type', 'attr', 'name', 'size'
  LINEAR_ATTRS = ('dram',)
  def __str__(self):
    return 'param {}{}: {}{}'.format(
      self.haoda_type, ''.join(map(', {}'.format, self.attr)),
      self.name, ''.join(map('[{}]'.format, self.size)))

class ParamAttr(ir.Node):
  SCALAR_ATTRS = 'dup', 'partitioning'
  def __str__(self):
    if self.dup is not None:
      return 'dup {}'.format(self.dup)
    result = 'partition {0.strategy}'.format(self.partitioning)
    if self.partitioning.strategy == 'cyclic':
      result += ' factor={}'.format(self.partitioning.factor)
    if self.partitioning.dim is not None:
      result += ' dim={}'.format(self.partitioning.dim)
    return result

class SodaProgram(ir.Node):
  SCALAR_ATTRS = ('border', 'burst_width', 'cluster', 'iterate', 'app_name',
                  'unroll_factor', 'input_stmts', 'param_stmts', 'local_stmts',
                  'output_stmts')
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    for node in self.input_stmts:
      if hasattr(self, 'tile_size'):
        # pylint: disable=access-member-before-definition
        if self.tile_size != node.tile_size:
          msg = ('tile size %s doesn\'t match previous one %s' %
               # pylint: disable=access-member-before-definition
               (node.tile_size, self.tile_size))
          raise util.SemanticError(msg)
      elif node.tile_size[:-1]:
        self.tile_size = node.tile_size
        self.dim = len(self.tile_size)
    # deal with 1D case
    if not hasattr(self, 'tile_size'):
      # pylint: disable=undefined-loop-variable
      self.tile_size = node.tile_size
      self.dim = len(self.tile_size)

  def __str__(self):
    return '\n'.join(filter(None, (
      'border: {}'.format(self.border),
      'burst width: {}'.format(self.burst_width),
      'cluster: {}'.format(self.cluster),
      'iterate: {}'.format(self.iterate),
      'kernel: {}'.format(self.app_name),
      'unroll factor: {}'.format(self.unroll_factor),
      '\n'.join(map(str, self.input_stmts)),
      '\n'.join(map(str, self.param_stmts)),
      '\n'.join(map(str, self.local_stmts)),
      '\n'.join(map(str, self.output_stmts)))))

CLASSES = (
  InputStmt,
  LocalStmt,
  OutputStmt,
  ir.Let,
  ir.Ref,
  ir.Expr,
  ir.LogicAnd,
  ir.BinaryOr,
  ir.Xor,
  ir.BinaryAnd,
  ir.EqCmp,
  ir.LtCmp,
  ir.AddSub,
  ir.MulDiv,
  ir.Unary,
  ir.Operand,
  ir.Cast,
  ir.Call,
  ir.Var,
  ParamStmt,
  ParamAttr,
  SodaProgram,
)



from typing import (
    Iterable,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import collections
import logging
import operator
import types

from haoda import ir
from soda import tensor
import soda.visitor

_logger = logging.getLogger().getChild(__name__)


def shift(obj, offset, excluded=(), op=operator.sub, verbose=False):
  """Shift soda.ir.Ref with the given offset.

  All soda.ir.Ref, excluding the given names, will be shifted with the
  given offset using the given operator. The operator will be applied pointwise
  on the original index and the given offset.

  Args:
    obj: A haoda.ir.Node or a tensor.Tensor object.
    offset: Second operand given to the operator.
    excluded: Sequence of names to be excluded from the mutation. Default to ().
    op: Shifting operator. Should be either add or sub. Default to sub.
    verbose: Whether to log shiftings. Default to False.
  Returns:
    Mutated obj. If obj is an IR node, it will be a different object than the
    input. If obj is a tensor, it will be the same object but with fields
    mutated.
  """
  if op not in (operator.add, operator.sub):
    _logger.warn('shifting with neither + nor -, which most likely is an error')

  def visitor(obj, args):
    if isinstance(obj, ir.Ref):
      if obj.name not in excluded:
        new_idx = tuple(op(a, b) for a, b in zip(obj.idx, offset))
        if verbose:
          _logger.debug('reference %s(%s) shifted to %s(%s)', obj.name,
                        ', '.join(map(str, obj.idx)), obj.name,
                        ', '.join(map(str, new_idx)))
        obj.idx = new_idx

  if isinstance(obj, ir.Node):
    return obj.visit(visitor)
  if isinstance(obj, tensor.Tensor):
    obj.mutate(visitor)
  else:
    raise TypeError('argument is not an IR node or a tensor')
  return obj


def normalize(obj: Union[ir.Node, Iterable[ir.Node]],
              references: Optional[Mapping[str, Tuple[int, ...]]] = None):
  """Make the least access index 0.

  Works on an ir.Node or an iterable of ir.Nodes. If it is shifted, a different
  object is constructed and returned. Otherwise, obj will be returned as-is.

  Args:
    obj: A node or an iterable of nodes.
  Returns:
    Normalized node or iterable.
  Raises:
    TypeError: If argument is not an ir.Node or an iterable of ir.Nodes.
  """
  if isinstance(obj, types.GeneratorType):
    return normalize(tuple(obj))
  norm_idx = soda.visitor.get_normalize_index(obj, references)
  shifter = lambda x: shift(x, norm_idx) if any(norm_idx) else x
  if isinstance(obj, collections.Iterable):
    return type(obj)(map(shifter, obj))  # type: ignore
  if isinstance(obj, ir.Node):
    return shifter(obj)
  raise TypeError('argument is not an ir.Node or an iterable of ir.Nodes')


NodeT = TypeVar('NodeT', bound=ir.Node)


def replace_expressions(
    obj: NodeT,
    cses: MutableMapping[NodeT, ir.Ref],
    used: Optional[MutableMapping[NodeT, NodeT]] = None,
    references: Optional[Mapping[str, Tuple[int, ...]]] = None,
) -> NodeT:
  """Get AST with common subexpression elimination.

  Get AST with the given common subexpressions. If used is not None, the used
  common subexpressions will be added to used.

  Args:
    obj: An ir.Node.
    cses: Dict mapping normalized common subexpressions to the new ir.Ref.
    used: Set of used common subexpressions, or None.
  Returns:
    The ir.Node as the AST.
  """

  def visitor(
      obj: NodeT,
      args: Tuple[MutableMapping[NodeT, ir.
                                 Ref], Optional[MutableMapping[NodeT, NodeT]]]
  ) -> NodeT:
    cses, used = args
    norm_idx = soda.visitor.get_normalize_index(obj, references)
    normalized = shift(obj, norm_idx) if any(norm_idx) else obj
    if normalized in cses:
      if used is not None:
        if normalized not in used:
          used[normalized] = replace_expressions(
              normalized, {k: v for k, v in cses.items() if k != normalized},
              used)
      new_obj = shift(cses[normalized], norm_idx, op=operator.add)
      _logger.debug('replacing %s with %s', obj, new_obj)
      return new_obj
    return obj

  return obj.visit(visitor, (cses, used))
